import uuid
from datetime import datetime, timedelta

from pytz import UTC

from django.test import Client, TestCase, override_settings
from django.urls import reverse

from core.models import Assembly, Conference, ConferenceMember, Event, PlatformUser

TEST_CONF_ID = uuid.uuid4()


@override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class EventsTestCase(TestCase):
    def setUp(self):
        self.conference1 = Conference(id=TEST_CONF_ID, slug='foo', name='Foo Conference', is_public=True)
        self.conference1.save()

        self.human_user = PlatformUser(username='bernd', user_type=PlatformUser.Type.HUMAN)
        self.human_user.save()
        self.human_cm = ConferenceMember(conference=self.conference1, user=self.human_user)
        self.human_cm.save()

        assembly = Assembly(conference=self.conference1, name='DUMMY', slug='dummy', state=Assembly.State.ACCEPTED, is_official=True)
        assembly.save()

        self.event = Event(
            conference=self.conference1,
            slug='event1',
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
            kind=Event.Kind.OFFICIAL,
        )
        self.event.save()

    def test_EventMyFavorites(self):
        c = Client()
        resp = c.get(reverse('api:my-events'))
        self.assertEqual(resp.status_code, 403)
        c.force_login(self.human_user)

        resp = c.get(reverse('api:my-events'))
        self.assertEqual(resp.json(), [])

        self.human_user.favorite_events.add(self.event)
        resp = c.get(reverse('api:my-events'))
        self.assertEqual(resp.json(), [str(self.event.pk)])

    def test_EventMyFavoritesSet_put(self):
        c = Client()
        resp = c.put(reverse('api:my-events-set', kwargs={'pk': str(uuid.uuid4())}))
        self.assertEqual(resp.status_code, 403)
        c.force_login(self.human_user)

        self.assertEqual(list(self.human_user.favorite_events.all()), [])
        resp = c.put(reverse('api:my-events-set', kwargs={'pk': str(self.event.pk)}))
        self.assertEqual(list(self.human_user.favorite_events.all()), [self.event])

    def test_EventMyFavoritesSet_delete(self):
        c = Client()
        resp = c.delete(reverse('api:my-events-set', kwargs={'pk': str(uuid.uuid4())}))
        self.assertEqual(resp.status_code, 403)
        c.force_login(self.human_user)

        self.human_user.favorite_events.add(self.event)

        self.assertEqual(list(self.human_user.favorite_events.all()), [self.event])
        resp = c.delete(reverse('api:my-events-set', kwargs={'pk': str(self.event.pk)}))
        self.assertEqual(list(self.human_user.favorite_events.all()), [])
