import io
import json
import zipfile

from rest_framework import generics, permissions
from rest_framework.decorators import api_view
from rest_framework.response import Response

from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponse
from django.views.generic import View

from core.choices import UserBadgeVisibility
from core.models.badges import Badge, UserBadge
from core.models.users import PlatformUser, UserTimelineEntry

from api.serializers import UserTimelineEntrySerializer


@api_view(['GET'])
def profile(request, format=None):  # pylint: disable=redefined-builtin
    u = request.user

    if not u.is_authenticated:
        u = PlatformUser.get_anonymous_user()

    return Response(
        {
            'authenticated': u.is_authenticated,
            'username': u.username,
            'flags': PlatformUser.get_user_flags(u),
        }
    )


@api_view(['GET'])
def friends(request, format=None):  # pylint: disable=redefined-builtin
    u = request.user
    if not u.is_authenticated:
        return Response([])

    return Response(
        [
            {
                'user': c.contact.username,
                'pending': c.pending,
                # status may only be shown if that is public or the contact has a allowed sharing to us
                'status': c.contact.status if c.contact.status_public or (c.reverse_contact_share_status(default=False) and not c.pending) else None,
                'receive_dms': c.receive_dms,
                'receive_dm_images': c.receive_dm_images,
            }
            for c in u.contacts.all()
        ]
    )


@api_view(['GET'])
def badges(request, format=None):  # pylint: disable=redefined-builtin
    u = request.user
    if not u.is_authenticated:
        return Response([])

    return Response(
        [
            {
                'conference': ub.badge.conference.name,
                'assembly': ub.badge.issuing_assembly.name,
                'name': ub.badge.name,
                'image_url': ub.badge.image.url,
                'is_achievement': ub.badge.is_achievement,
            }
            for ub in u.badges.filter(visibility=UserBadgeVisibility.PUBLIC)
        ]
    )


class BadgeExportView(LoginRequiredMixin, View):
    def get(self, *args, **kwargs):
        badges = UserBadge.objects.filter(user=self.request.user).select_related('badge', 'badge__issuing_assembly')

        fs = io.BytesIO()
        archive = zipfile.ZipFile(fs, mode='w')

        data = []

        for ub in badges:
            entry = {
                'id': str(ub.badge.id),
                'title': ub.badge.name,
                'hidden': ub.badge.state == Badge.State.HIDDEN,
                'accepted': ub.accepted_by_user,
                'visibility': ub.visibility,
                'assembly': ub.badge.issuing_assembly.name,
            }
            data.append(entry)

            try:
                image_filename = f'badge-{ub.badge_id}' + ('.jpg' if ub.badge.image.name.lower().split('.')[-1] in ['jpg', 'jpeg'] else '.png')
                badge_image = ub.badge.image.read()
                with archive.open(image_filename, 'w') as export_file:
                    export_file.write(badge_image)
                entry['image'] = image_filename
            except Exception:
                entry['image'] = None

        with archive.open('badges.json', 'w') as data_file:
            data_file.write(json.dumps(data, indent=2).encode('utf-8'))

        archive.close()

        response = HttpResponse(fs.getvalue(), content_type='archive/zip')
        response['Content-Disposition'] = f'attachment; filename="badges-{self.request.user.username}.zip"'
        return response


class UserTimelineList(generics.ListAPIView):
    serializer_class = UserTimelineEntrySerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        return UserTimelineEntry.objects.filter(user=self.request.user)
