import json
import uuid
from datetime import datetime, timedelta

from django.conf import settings
from django.template.defaulttags import register
from django.utils.crypto import get_random_string
from django.utils.timezone import get_current_timezone
from django.utils.translation import get_language

from core.models.assemblies import Assembly


@register.filter
def get_item(dictionary, key):
    assert isinstance(dictionary, dict)
    return dictionary.get(key)


@register.filter
def get_item_or_original(dictionary, key):
    return dictionary.get(key) if isinstance(dictionary, dict) else dictionary


@register.filter
def get_customfield(assembly, key):
    assert isinstance(assembly, Assembly)
    return assembly.get_field(key)


@register.filter
def get_language_item(dictionary):
    if not isinstance(dictionary, dict):
        return dictionary

    lang = get_language()
    if lang in dictionary:
        return dictionary[lang]

    for lang in settings.LANGUAGES:
        if lang in dictionary:
            return dictionary[lang]

    return dictionary[0]


@register.filter
def naturaltimespan(x: timedelta) -> str or None:
    if x is None:
        return ''
    if not isinstance(x, timedelta):
        raise ValueError('Expected a timedelta object.')

    seconds = x.total_seconds()

    minutes = int(seconds / 60)
    seconds -= minutes * 60

    hours = int(minutes / 60)
    minutes -= hours * 60

    result = []
    if hours > 0:
        result.append(f'{hours}h')
    if minutes > 0:
        result.append(f'{minutes}m')
    if seconds > 0:
        result.append(f'{seconds}s')

    return ' '.join(result) if result else None


@register.filter
def from_unixtimestamp(value):
    return datetime.fromtimestamp(int(value), get_current_timezone())


@register.filter
def is_dict(value):
    return isinstance(value, dict)


@register.filter
def filter_dict(value, key, *additional_keys):
    if not isinstance(value, dict):
        return value

    keys = list(additional_keys)
    keys.append(key)
    return {k: v for k, v in value.items() if k not in keys}


@register.filter
def json_indent(value):
    return json.dumps(value, indent=2)


@register.inclusion_tag('backoffice/_disclaimer_block.html', takes_context=True)
def show_disclaimer(context, *args, **kwargs):
    key = args[0] if args else kwargs.get('key')
    if key is None:
        raise Exception('No key has been given.')

    conf = context.get('conference')
    if conf is None:
        raise Exception('There is no conference in the current context.')

    if not conf.has_disclaimer(key):
        return {'disclaimer': None}

    disclaimer = conf.get_disclaimer(key)
    return {'disclaimer': disclaimer, 'disclaimer_type': kwargs.get('type')}


@register.filter(name='split')
def split(value, separator):
    return value.split(separator)


@register.filter(name='ensure_uuid')
def ensure_uuid(value):
    if isinstance(value, uuid.UUID):
        return value

    if not value:
        return None

    return uuid.UUID(value)


@register.simple_tag
def gen_rand_str(length: int = 8):
    return get_random_string(length=length)
