import logging
from typing import Any

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.forms.forms import BaseForm
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormView, UpdateView

from core.forms import ConferencePublicationForm, ConferenceRegistrationForm
from core.models import Conference, ConferenceMember

from backoffice.forms import ConferenceSelectionForm
from backoffice.views.mixins import ConferenceRequiredMixin, LoginRequiredMixin

logger = logging.getLogger(__name__)


class ConferenceSelectionView(LoginRequiredMixin, FormView):
    """
    A class-based view that will let you select the current :model:`core.Conference` for you session.

    It will also be used if no conference is available and show an error message to the user.

    **Context:**

    ``conferences``
        A list of all available :model:`core.Conference` for the current user.

    **Template:**

    :template:`backoffice/conferences/selection.html`

    """

    form_class = ConferenceSelectionForm
    template_name = 'backoffice/conferences/selection.html'
    active_page = 'conferences'

    def form_valid(self, form: ConferenceSelectionForm):
        """Sets the selected conference for the session if the form is valid.

        Args:
            form (ConferenceForm): A form containing the selected conference

        Returns:
            Response: A redirection to the backoffice index
        """
        conf_slug = form.cleaned_data['conference_slug']
        conference = Conference.objects.accessible_by_user(self.request.user).get(slug=conf_slug)
        self.request.session['conference'] = conference.slug
        return redirect('backoffice:index')

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            'conferences': Conference.objects.accessible_by_user(self.request.user),
            'active_page': self.active_page,
        }


class ConferenceFormMixin(ConferenceRequiredMixin, FormView):
    object: Conference | None
    model = Conference
    context_object_name = 'conference'
    permission_required = ['core.conference_admin', 'core.assembly_registration_admin']
    require_all_permissions = False
    require_staff = True
    form_permissions_required = []
    sidebar_caption = _('nav_conference')
    active_page: str

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            'has_form_perms': self.request.user.has_conference_staff_permission(self.conference, *self.form_permissions_required),
            'active_page': self.active_page,
            'sidebar': {
                'title': self.sidebar_caption,
                'items': [
                    {
                        'caption': _('Conference__Publication'),
                        'link': reverse(
                            'backoffice:conference-publication',
                            kwargs={'pk': self.conference.pk},
                        ),
                    },
                    {
                        'caption': _('Conference__Registration'),
                        'link': reverse(
                            'backoffice:conference-registration',
                            kwargs={'pk': self.conference.pk},
                        ),
                    },
                ],
            },
        }

    def get_form(self, form_class: type | None = None) -> ConferencePublicationForm | BaseForm:
        form = super().get_form(form_class)
        member = ConferenceMember.objects.filter(user=self.request.user, conference=self.object).first()
        if member and member.has_perms(*self.form_permissions_required, require_staff=self.require_staff):
            return form
        for field in form.fields:
            form.fields[field].disabled = True
        return form

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        if not self.request.user.has_conference_staff_permission(self.conference, *self.form_permissions_required):
            raise PermissionDenied
        return super().post(request, *args, **kwargs)


class ConferencePublicationView(ConferenceFormMixin, UpdateView):
    """
    An UpdateView to change the publication settings of the current :model:`core.conference`.
    """

    form_class = ConferencePublicationForm
    template_name = 'backoffice/conferences/publication_edit.html'
    active_page = 'conference'
    form_permissions_required = ['core.conference_admin']

    def get_form_kwargs(self, *args, **kwargs) -> dict[str, Any]:
        publish = False
        if self.object:
            publish = (not self.object.is_public and self.request.GET.get('publish') == 'true') or (
                self.object.is_public and self.request.GET.get('recall') == 'true'
            )
        return {
            **super().get_form_kwargs(*args, **kwargs),
            'publish': publish,
        }

    def form_valid(self, form: ConferencePublicationForm):
        # Try to safe the form before creating logs and messages
        self.object = form.save()
        logger.info(
            'changed publication info of "%(conference)s" by %(user)s',
            {
                'conference': form.instance,
                'user': self.request.user,
            },
        )

        if form.publish and self.object.is_public:
            messages.success(self.request, _('Conference__publish__success'))
        elif form.publish and not self.object.is_public:
            messages.success(self.request, _('Conference__recall__success'))
        elif form.create:
            messages.success(self.request, _('Conference__create__success'))
        else:
            messages.success(self.request, _('Conference__update__success'))

        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form: ConferencePublicationForm):
        if form.publish:
            form.instance.is_public = False
            messages.error(self.request, _('Conference__publish__failed'))
        else:
            messages.error(self.request, _('Conference__form__invalid'))

        return super().form_invalid(form)

    def get_success_url(self):
        return reverse('backoffice:conference-publication', kwargs={'pk': self.object.pk})


class ConferenceRegistrationView(ConferenceFormMixin, UpdateView):
    """
    An UpdateView to change to current :model:`core.conference` settings.
    """

    template_name = 'backoffice/conferences/registration_edit.html'
    form_class = ConferenceRegistrationForm
    active_page = 'conference'
    form_permissions_required = ['core.assembly_registration_admin']

    def form_valid(self, form: ConferencePublicationForm):
        # Try to safe the form before creating logs and messages
        self.object = form.save()
        logger.info(
            'changed publication info of "%(conference)s" by %(user)s',
            {
                'conference': form.instance,
                'user': self.request.user,
            },
        )
        messages.success(self.request, _('Conference__registration_update__success'))

        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return reverse('backoffice:conference-registration', kwargs={'pk': self.object.pk})

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            'assembly_team_from_email': settings.ASSEMBLY_TEAM_FROM_EMAIL,
            'assembly_team_reply_to': ', '.join(settings.ASSEMBLY_TEAM_REPLY_TO),
        }
