import logging

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import ModelFormMixin

from core.models.activitylog import ActivityLogChange, ActivityLogEntry, model_form_valid_activitylog
from core.models.conference import ConferenceExportCache
from core.models.events import Event, EventParticipant

from backoffice.forms import EventForm
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class EventFormMixin(ModelFormMixin):
    model = Event
    form_class = EventForm
    event_type_name = _('event')
    is_sos = False

    def form_invalid(self, form):
        if form.publish:
            form.instance.is_public = False
            messages.error(self.request, _('Event__publish__failed %(event_type)s') % {'event_type': self.event_type_name})
        else:
            messages.error(self.request, _('Event__form__invalid %(event_type)s') % {'event_type': self.event_type_name})
        return super().form_invalid(form)

    @model_form_valid_activitylog(context='backoffice')
    def form_valid(self, form):
        result = super().form_valid(form)
        if form.publish and form.instance.is_public:
            form.instance.log_activity(
                user=self.request.user,
                kind=ActivityLogEntry.Kind.SYSTEM,
                comment='PUBLISHED',
                context='backoffice.events',
                is_public=ActivityLogChange(new='yes', old='no'),
            )
            messages.success(
                self.request, _('Event__published %(event_id)s %(event_type)s') % {'event_id': form.instance.id, 'event_type': self.event_type_name}
            )
        elif form.publish and not form.instance.is_public:
            form.instance.log_activity(
                user=self.request.user,
                kind=ActivityLogEntry.Kind.SYSTEM,
                comment='RECALLED',
                context='backoffice.events',
                is_public=ActivityLogChange(new='no', old='yes'),
            )
            messages.success(
                self.request, _('Event__recalled %(event_id)s %(event_type)s') % {'event_id': form.instance.id, 'event_type': self.event_type_name}
            )
        elif form.create:
            form.instance.participants.create(
                role=EventParticipant.Role.SPEAKER,
                participant=self.request.user,
                is_accepted=True,
                is_public=True,
            )
            form.instance.log_activity(
                user=self.request.user,
                kind=ActivityLogEntry.Kind.SYSTEM,
                context='backoffice.events',
                comment='add creator as speaker',
                **{
                    'participants.added': ActivityLogChange(new=self.request.user.username),
                    f'participants[{self.request.user.username}].role': ActivityLogChange(new=EventParticipant.Role.SPEAKER),
                    f'participants[{self.request.user.username}].is_public': ActivityLogChange(new='yes'),
                    f'participants[{self.request.user.username}].is_accepted': ActivityLogChange(new='yes'),
                },
            )
            messages.success(self.request, _('Event__created %(event_id)s %(event_type)s') % {'event_id': form.instance.id, 'event_type': self.event_type_name})
        else:
            messages.success(self.request, _('Event__updated %(event_id)s %(event_type)s') % {'event_id': form.instance.id, 'event_type': self.event_type_name})
            ConferenceExportCache.signal_schedule_modification(self.conference)
        return result

    def get_form_kwargs(self, *args, **kwargs):
        return {'conference': self.conference, **super().get_form_kwargs(*args, **kwargs)}


class EventPublicationMixin(ModelFormMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        try:
            self.get_object().clean(True)
        except ValidationError as error:
            context.update({'publication_errors': error.message_dict})
        return context

    def get_form_kwargs(self, *args, **kwargs):
        publish = (not self.object.is_public and self.request.GET.get('publish') == 'true') or (
            self.object.is_public and self.request.GET.get('recall') == 'true'
        )
        return {'publish': publish, **super().get_form_kwargs(*args, **kwargs)}


class AssemblyEventFormMixin(AssemblyMixin, EventFormMixin):
    template_name = 'backoffice/assembly_event.html'

    assembly_url_param = 'assembly'
    assembly_management = True

    def get_form_kwargs(self, *args, **kwargs):
        return {'assembly': self.assembly, **super().get_form_kwargs(*args, **kwargs)}

    def get_success_url(self):
        return reverse('backoffice:assembly-event', kwargs={'assembly': self.assembly.id, 'pk': self.object.id})


class SoSFormMixin(EventFormMixin):
    template_name = 'backoffice/sos_create_edit.html'
    event_type_name = _('SoS')
    is_sos = True

    def get_context_data(self, *args, **kwargs):
        return {
            'active_page': 'sos',
            **super().get_context_data(*args, **kwargs),
        }

    def get_form_kwargs(self, *args, **kwargs):
        return {'owner': self.request.user, **super().get_form_kwargs(*args, **kwargs)}

    def get_success_url(self):
        return reverse('backoffice:sos-edit', kwargs={'pk': self.object.id})
