import logging
from typing import Any

from django.contrib import messages
from django.http import HttpRequest, HttpResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView, DeleteView, UpdateView

from core.models.assemblies import Assembly
from core.models.projects import Project

from backoffice.views.mixins import ConferenceLoginRequiredMixin
from backoffice.views.projects.mixins import ProjectListViewMixin, SoProjectFormMixin

logger = logging.getLogger(__name__)


class SelfOrganizedProjectListView(ProjectListViewMixin):
    def get_queryset(self):
        return Project.objects.manageable_by_user(self.conference, user=self.request.user).filter(assembly__isnull=True)


class ProjectCreateView(SoProjectFormMixin, CreateView):
    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        self.object = None
        return super().post(request, *args, **kwargs)


class SelfOrganizedProjectUpdateView(SoProjectFormMixin, UpdateView):
    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)


class SelfOrganizedProjectDeleteView(ConferenceLoginRequiredMixin, DeleteView):
    assembly: Assembly | None = None

    def get_success_url(self) -> str:
        if self.assembly is not None:
            return reverse('backoffice:assembly-projects', kwargs={'assembly': self.assembly.id})
        return reverse('backoffice:sos')

    def get_queryset(self):
        return Project.objects.manageable_by_user(self.conference, user=self.request.user)

    def get_log_data(self, form) -> dict[str, str]:
        project = self.get_object()
        self.assembly = project.assembly
        log_data = {
            'project_type': _('AssemblyProject') if project.assembly else _('SelfOrganizedProject'),
            'project': project.name,
            'user': self.request.user,
            'owner': 'owner ' if project.owner else '',
            'assembly': f' on {project.assembly} ' if project.assembly else '',
        }
        return log_data

    def form_valid(self, form):
        log_data = self.get_log_data(form)
        response = super().form_valid(form)
        messages.success(self.request, _('Project__removed__success %(project_type)s %(project)s') % log_data)
        logger.info('removed %(project_type)s "%(project)s" %(assembly)sby %(owner)s%(user)s', log_data)
        return response

    def form_invalid(self, form):
        log_data = self.get_log_data(form)
        response = super().form_invalid(form)
        messages.error(self.request, _('Project__removal_failure %(project_type)s %(project)s') % log_data)
        logger.info('removed %(project_type)s "%(project)s" %(assembly)sby %(owner)s%(user)s', log_data)
        return response
