from contextlib import suppress
from uuid import UUID

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db.models import QuerySet
from django.forms import HiddenInput, ModelForm, Select
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from core.models import (
    Assembly,
    Conference,
    Invitation,
    PlatformUser,
    Team,
)


class InvitationBaseForm(ModelForm):
    class Meta:
        model = Invitation
        fields = (
            'requester_id',
            'requested_id',
            'comment',
        )

    requester_id: str | None = None
    requester_link: str | None = None
    requested_id: str | None = None
    requested_link: str | None = None

    def __init__(
        self,
        *args,
        conference: Conference,
        user: PlatformUser,
        instance: Invitation | None,
        requester_id: str | None = None,
        requested_id: str | None = None,
        **kwargs,
    ):
        kwargs['initial'] = kwargs.get('initial', {})

        self.conference = conference
        self.user = user

        self.create = instance is None
        if instance is None:
            if not requester_id:  # pragma: no cover
                raise ValueError('Requester ID is required')
            self.requester_id = str(requester_id)
            self.requested_id = str(requested_id) if requested_id else None
        else:
            self.requester_id = str(instance.requester_id)
            self.requested_id = str(instance.requested_id)

        kwargs['initial']['requester_id'] = self.requester_id
        if 'data' in kwargs:
            kwargs['data'] = kwargs['data'].copy()
            kwargs['data']['requester_id'] = self.requester_id
        if self.requested_id is not None:
            kwargs['initial']['requested_id'] = self.requested_id
            if 'data' in kwargs:
                kwargs['data'] = kwargs['data'].copy()
                kwargs['data']['requested_id'] = self.requested_id

        super().__init__(*args, instance=instance, **kwargs)
        # self.fields is only available after super().__init__ is called
        self.fields['requester_id'].widget.attrs['disabled'] = True
        if self.requested_id is not None:
            self.fields['requested_id'].widget.attrs['disabled'] = True
        if instance:
            self.permissions = {
                'requester': self.user.has_perm('core.change_invitation', self.instance),
                'requested': self.user.has_perm('core.accept_invitation', self.instance),
            }

    def save(self, commit: bool = True) -> Invitation:
        invitation = super().save(commit=False)
        if self.create:
            invitation.conference = self.conference
            invitation.sender = self.user
        if commit:  # pragma: no cover
            invitation.save()
        return invitation

    def clean_requester_id(self):
        data = self.cleaned_data['requester_id']
        if not self.create and self.instance and str(self.instance.requester_id) != str(data):
            raise ValidationError(_('Invitation__error__requester_change'))
        return data

    def clean_requested_id(self):
        data = self.cleaned_data['requested_id']
        if not self.create and self.instance and str(self.instance.requested_id) != str(data):
            raise ValidationError(_('Invitation__error__requested_change'))
        return data


class InvitationHabitatForm(InvitationBaseForm):
    class Meta(InvitationBaseForm.Meta):
        widgets = {
            'requester_id': Select(),
            'requested_id': Select(),
        }

    def get_queryset(self) -> QuerySet[Assembly]:
        return Assembly.objects.associated_with_user(conference=self.conference, user=self.user, show_public=True, staff_can_see=False)

    def get_choices(self, *, habitat: bool) -> list[tuple[str, str]]:
        if habitat:
            return self.get_queryset().exclude(hierarchy=Assembly.Hierarchy.REGULAR).values_list('id', 'name')
        else:
            return self.get_queryset().filter(hierarchy=Assembly.Hierarchy.REGULAR).values_list('id', 'name')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.create:
            assembly = get_object_or_404(Assembly, pk=self.requester_id)
            habitat = assembly.hierarchy != Assembly.Hierarchy.REGULAR
            self.fields['requester_id'].widget.choices = self.get_choices(habitat=habitat)
            self.fields['requested_id'].widget.choices = self.get_choices(habitat=(not habitat))
        else:
            habitat = self.instance.requester.hierarchy != Assembly.Hierarchy.REGULAR
            self.requester_link = reverse_lazy('backoffice:assembly', kwargs={'pk': self.instance.requester.pk})
            self.requested_link = reverse_lazy('backoffice:assembly', kwargs={'pk': self.instance.requester.pk})
        self.fields['requester_id'].label = _('Assembly__parent') if habitat else _('Assembly')
        self.fields['requested_id'].label = _('Assembly') if habitat else _('Assembly__parent')

    def save(self, commit: bool = True) -> Invitation:
        invitation = super().save(commit=False)
        assembly_type = ContentType.objects.get_for_model(Assembly)
        invitation.type = Invitation.InvitationType.HABITAT
        invitation.requester_type = assembly_type
        invitation.requested_type = assembly_type
        if commit:  # pragma: no branch
            invitation.save()
        return invitation


class InvitationTeamForm(InvitationBaseForm):
    class Meta(InvitationBaseForm.Meta):
        widgets = {
            'requester_id': HiddenInput(),
        }

    invitation_team: Team | None = None
    invitation_user: PlatformUser | None = None

    def __init__(self, *args, instance: Invitation | None, from_team: bool = False, **kwargs):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        super().__init__(*args, instance=instance, **kwargs)
        if instance:  # We are not in a create form
            self.requester_id = str(instance.requester.uuid)
            self.requested_id = str(instance.requested.uuid)
            from_team = instance.requester_type.model == 'team'
            self.invitation_team = instance.requester if from_team else instance.requested
            self.invitation_user = instance.requested if from_team else instance.requester
            assert self.invitation_team is not None
            assert self.invitation_user is not None
        elif from_team:
            self.invitation_team = get_object_or_404(Team, uuid=self.requester_id)
            if self.requested_id:
                self.invitation_user = get_object_or_404(PlatformUser, uuid=self.requested_id)
        else:
            self.invitation_user = get_object_or_404(PlatformUser, uuid=self.requester_id)
            if self.requested_id:
                self.invitation_team = get_object_or_404(Team, uuid=self.requested_id)
        self.from_team = from_team

        self.fields['requester_id'].label = _('Team') if from_team else _('TeamMember')
        self.fields['requester_id'].widget = HiddenInput()
        self.requester_text = self.invitation_team.name if from_team else self.invitation_user.name
        self.requester_link = (
            reverse_lazy('backoffice:team', kwargs={'uuid': self.invitation_team.uuid})
            if from_team
            else hub_absolute('plainui:user', user_slug=self.invitation_user.slug)
        )

        self.fields['requested_id'].label = _('TeamMember') if from_team else _('Team')
        if self.requested_id:
            self.fields['requested_id'].widget = HiddenInput()
            self.requested_text = self.invitation_user.name if from_team else self.invitation_team.name
            self.requested_link = (
                hub_absolute('plainui:user', user_slug=self.invitation_user.slug)
                if from_team
                else reverse_lazy('backoffice:team', kwargs={'uuid': self.invitation_team.uuid})
            )

    def clean_requested_id(self):
        requested_query = self.cleaned_data['requested_id']

        requested_id = None

        with suppress(ValueError):
            requested_id = get_object_or_404(PlatformUser if self.from_team else Team, uuid=UUID(requested_query, version=4)).pk

        if not requested_id:
            with suppress(ValueError):
                requested_id = int(requested_query)

        if not requested_id and self.from_team:
            requested_id = get_object_or_404(PlatformUser, username__iexact=requested_query).pk
        if not requested_id:
            requested_id = get_object_or_404(Team, name__iexact=requested_query).pk

        self.cleaned_data['requested_id'] = requested_id
        return super().clean_requested_id()

    def clean_requester_id(self):
        self.cleaned_data['requester_id'] = self.invitation_team.pk if self.from_team else self.user.pk
        return super().clean_requester_id()

    def save(self, commit: bool = True) -> Invitation:
        invitation = super().save(commit=False)
        types = ContentType.objects.get_for_models(Team, PlatformUser)

        invitation.type = Invitation.InvitationType.TEAM_TO_MEMBER if self.from_team else Invitation.InvitationType.MEMBER_TO_TEAM
        invitation.requester_type = types[Team] if self.from_team else types[PlatformUser]
        invitation.requested_type = types[PlatformUser] if self.from_team else types[Team]
        if self.requested_id:
            invitation.requested_id = get_object_or_404(PlatformUser if self.from_team else Team, uuid=self.requested_id).pk
        if commit:  # pragma: no branch
            invitation.save()
        return invitation
