import sys
import time
from os import getenv

from django.core.management.base import CommandError
from django.db.models import Max
from django.utils import timezone
from django_rich.management import RichCommand

from core.models.conference import Conference, ConferenceExportCache
from core.models.messages import DirectMessage
from core.models.pages import StaticPageNamespace
from core.models.schedules import ScheduleSource, ScheduleSourceImport
from core.models.voucher import Voucher


def _no_traceback_exception_hook(exc_type, exc_val, traceback):
    """Hook to prevent traceback on KeyboardInterrupt."""


class Command(RichCommand):
    DEFAULT_FOREVER_INTERVAL = 120

    def add_arguments(self, parser):
        parser.add_argument('--forever', action='store_true', help='repeat the housekeeping forever (until Ctrl+C is pressed)')
        parser.add_argument(
            '--forever-interval',
            type=int,
            default=None,
            help='seconds to wait between housekeeping runs (can be set via FOREVER_INTERVAL environment variable, too)',
        )
        parser.add_argument('--skip-schedule-imports', action='store_true', help="don't do schedule imports")
        parser.add_argument('--skip-wiki-imports', action='store_true', help="don't import wiki namespaces from upstream")

    def _housekeeping_directmessages(self):
        # clear all direct messages which are after their expiry date
        self.console.print('Deleting messages ... ', end='', style='red')
        deleted_msgs_count, _ = DirectMessage.objects.filter(autodelete_after__isnull=False, autodelete_after__lte=timezone.now()).delete()
        self.console.print(deleted_msgs_count)

    def _housekeeping_vouchers(self):
        # do auto-assignments
        self.console.print('Auto-assigning vouchers ... ', end='', style='green')
        vouchers_assigned = Voucher.do_auto_assignments()
        self.console.print(vouchers_assigned)

    def _housekeeping_scheduleimports(self):
        # schedules
        self.console.print('Schedule imports ... ', end='', style='blue')
        schedule_results = {}
        schedule_failure, schedule_success, schedule_skipped = 0, 0, 0
        for schedule in ScheduleSource.objects.all():
            # skip schedules which aren't due yet
            if not schedule.is_due:
                continue

            # skip schedules where an import is already running
            if schedule.has_running_import:
                schedule_results[str(schedule)] = None
                schedule_skipped += 1
                continue

            # create import job
            job = ScheduleSourceImport(schedule_source=schedule, state=ScheduleSourceImport.State.PREPARED)
            job.save()

            # execute job
            try:
                result = job.do_import()
                schedule_results[str(schedule)] = job.summary
                if result:
                    schedule_success += 1
                else:
                    schedule_failure += 1
            except Exception as err:
                schedule_results[str(schedule)] = 'EXCEPTION: ' + str(err)
                schedule_failure += 1

        # print schedule import results
        self.console.print(f'{schedule_success}/{len(schedule_results)} ({schedule_skipped} skipped)', style='cyan bold')
        for k, v in schedule_results.items():
            print('  ', k, ' => ', v, sep='')

    def _housekeeping_exportcache(self):
        ref = {}
        for c in Conference.objects.all():
            ts = []
            ts.append(c.tracks.aggregate(Max('last_update'))['last_update__max'])
            ts.append(c.events.aggregate(Max('last_update'))['last_update__max'])
            ts.append(c.rooms.aggregate(Max('last_update'))['last_update__max'])
            ref[c.id] = max([x for x in ts if x is not None], default=None)

        changed, total = 0, 0
        for entry in ConferenceExportCache.objects.all():
            total += 1

            # entries already marked for regeneration don't need further action
            if entry.needs_regeneration:
                continue

            # ignore cache entries with a newer generation entry
            ref_ts = ref[entry.conference_id]
            if ref_ts is not None and entry.last_generated > ref_ts:
                continue

            # flag this entry for regeneration
            entry.needs_regeneration = True
            entry.save(update_fields=['needs_regeneration'])
            changed += 1

        self.console.print(f'Flagged {changed} out of {total} cached exports for regeneration.', style='yellow')

    def _housekeeping_wikimports(self):
        for c in Conference.objects.all():
            for ns in StaticPageNamespace.objects.filter(conference=c).exclude(upstream_url__isnull=True).all():
                self.console.print(f'Importing "{c.slug}" wiki namespace "{ns.prefix}" ...', style='blue')
                ns.fetch_upstream()

    def handle(self, *args, **options):
        # call _do_housekeeping repeatedly (unless --forever is not set)
        forever = options.get('forever')
        forever_interval = int(options.get('forever_interval') or getenv('FOREVER_INTERVAL') or '0')
        if forever:
            if forever_interval <= 0:
                raise CommandError('The --forever-interval value must a positive value (in seconds).')
            self.console.print(f'Running housekeeping forever each {forever_interval}s:', end='\n\n')

        try:
            while True:
                if forever:
                    self.console.print(timezone.now().isoformat(), style='orange1', highlight=False)

                # call the individual housekeeping methods
                self._housekeeping_directmessages()
                self._housekeeping_vouchers()
                if not options.get('skip_schedule_imports'):
                    self._housekeeping_scheduleimports()
                self._housekeeping_exportcache()
                if not options.get('skip_wiki_imports'):
                    self._housekeeping_wikimports()

                if forever:
                    print()  # empty line
                    time.sleep(forever_interval)
                else:
                    break
        except KeyboardInterrupt:
            if sys.excepthook is sys.__excepthook__:
                sys.excepthook = _no_traceback_exception_hook
            print('Aborted.')
            raise
