from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from core.utils import resolve_internal_url, resolve_link


class Link(models.Model):
    class Meta:
        unique_together = [['content_type', 'object_id', 'name', 'link', 'link_type']]

    class LinkType(models.TextChoices):
        WEBSITE = 'website', _('RoomLink__type-website')
        CHAT = 'chat', _('RoomLink__type-chat')
        BBB = 'bbb', _('RoomLink__type-bbb')
        JITSI = 'jitsi', _('RoomLink__type-jitsi')
        PAD = 'pad', _('RoomLink__type-pad')
        MEDIA_CCC_DE = 'media.ccc.de', _('RoomLink__type-media_ccc_de')
        VIDEO = 'video', _('RoomLink__type-video')
        AUDIO = 'audio', _('RoomLink__type-audio')
        NAV = 'nav', _('RoomLink__type-nav')

    URL_LINK_TYPES = [LinkType.WEBSITE, LinkType.CHAT, LinkType.BBB, LinkType.JITSI, LinkType.PAD, LinkType.VIDEO, LinkType.AUDIO, LinkType.NAV]
    """All LinkTypes which require the link to be a valid URL. The notable exception is media_ccc_de where only a global id is required."""

    content_types = models.Q(app_label='core', model='project')
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, limit_choices_to=content_types)
    object_id = models.UUIDField()
    content_object = GenericForeignKey('content_type', 'object_id')
    name = models.CharField(max_length=200)

    link_type = models.CharField(max_length=20, choices=LinkType.choices, help_text=_('RoomLink__type__help'), verbose_name=_('RoomLink__type'))
    link = models.CharField(max_length=255)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._resolved_url = None

    def _cache_resolved_url_if_necessary(self):
        if self._resolved_url:
            return
        self._resolved_url = resolve_link(self.link)

    @property
    def resolved_url(self) -> str:
        self._cache_resolved_url_if_necessary()
        return self._resolved_url[1]

    @property
    def resolved_url_internal(self) -> bool:
        self._cache_resolved_url_if_necessary()
        return self._resolved_url[0] == 'internal'

    def clean(self):
        if self.link_type in Link.URL_LINK_TYPES:
            validator = URLValidator()
            try:
                validator(self.link)
            except ValidationError:
                raise ValidationError({'link': _('Link__link__must_be_url')})

    def to_dict(self):
        return {
            'type': self.link_type,
            'name': self.name,
            'uri': self.link,
            'url': resolve_internal_url(self.link),
        }

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        repr_str = f'<{self.__class__.__name__}: {self.name} ({self.link_type})'
        repr_str += f': {self.link}>' if self.link != self.name else '>'
        return repr_str
