import logging

import jwt

from django.conf import settings
from django.db import models
from django.db.utils import IntegrityError
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.models.conference import Conference
from core.models.users import PlatformUser

logger = logging.getLogger(__name__)


class TicketValidationError(Exception):
    pass


class ConferenceMemberTicket(models.Model):
    conference = ConferenceReference(related_name='tickets', on_delete=models.CASCADE)
    user = models.ForeignKey(PlatformUser, blank=True, null=True, related_name='tickets', on_delete=models.SET_NULL)
    ident = models.CharField(max_length=100, db_index=True)
    data = models.JSONField(blank=True, null=True)

    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(fields=['conference', 'user'], name='unique_conferencememberticket_conference_user'),
        ]

    @classmethod
    def validate_pretix_ticket(cls, conference: Conference, user: PlatformUser, token: str, validate_only=False):
        """
        Validates an incoming JWT from Pretix which must contain at least two fields in its payload:
        'uid' which is an arbitrary ticket-identifying character string, and
        'aud' which must match the conference's slug field.

        Rough/Example token content:
        ```json
        {
            "aud": "rc3",
            "exp": someexpirationtimestamp,
            "iat": issuingtimestamp,
            "iss": "tickets.events.ccc.de",
            "uid": "anonyme ticketid (alphanumerisch)"
        }
        ```

        If everything is fine (matching conference, ticket identifier has not been used before, etc.)
        the function returns the DB instance of the stored ticket in case registration needs to abort at a later point in time.
        Otherwise, i.e. if any validation step failed, an exception is raised.
        """

        assert conference is not None, 'Cannot validate a ticket without a Conference context.'
        assert user is not None, 'Cannot validate a ticket without knowing the user its meant for.'
        assert settings.PRETIX_SECRET_KEY is not None, 'PRETIX_SECRET_KEY has not been specified in the settings.'

        try:
            ticket_data = jwt.decode(
                token,
                settings.PRETIX_SECRET_KEY,
                algorithms=['HS256', 'HS512', 'ES256', 'ES512'],
                audience=conference.slug,
                issuer=settings.PRETIX_ISSUER,
            )

        except jwt.InvalidAudienceError as err:
            logger.error('JWT from pretix is for another audience: %s', err)
            raise TicketValidationError(_('ConferenceMemberTicket__token_wrong_conference'))

        except jwt.InvalidIssuerError as err:
            logger.error('Got JWT with unexpected issuer: %s', err)
            raise TicketValidationError(_('ConferenceMemberTicket__token_error'))

        except jwt.InvalidSignatureError:
            logger.error('Got invalid JWT from Pretix: %s', token)
            raise TicketValidationError(_('ConferenceMemberTicket__token_error'))

        except jwt.ExpiredSignatureError:
            logger.warning('Got expired JWT from Pretix: %s', token)
            raise TicketValidationError(_('ConferenceMemberTicket__token_error'))

        except jwt.DecodeError:
            raise TicketValidationError(_('ConferenceMemberTicket__token_error'))

        # read pretix's anonymized user id (and fail if field is missing or empty)
        try:
            pretix_ident = ticket_data.pop('uid').strip()
            assert pretix_ident != ''

        except (AssertionError, KeyError):
            logger.error('Pretix JWT is missing "uid": %s', ticket_data)
            raise TicketValidationError(_('ConferenceMemberTicket__token_missing_ident'))

        if validate_only:
            return (pretix_ident, ticket_data)

        # ensure this identifier has not been used before
        if user.is_authenticated and cls.objects.filter(conference=conference, user=user).exists():
            logger.warning('User already has a ticket: %s', pretix_ident)
            raise TicketValidationError(_('ConferenceMemberTicket__user_has_ticket_already'))

        # ensure this identifier has not been used before
        if cls.objects.filter(conference=conference, ident=pretix_ident).exists():
            logger.warning('Pretix ident has been used before: %s', pretix_ident)
            raise TicketValidationError(_('ConferenceMemberTicket__token_already_used'))

        if not user.is_authenticated:
            return None, None

        return (pretix_ident, ticket_data)

    @classmethod
    def redeem_pretix_ticket(cls, conference: Conference, user: PlatformUser, token: str):
        assert user.is_authenticated, "Anonymous Users can't join a conference."

        pretix_ident, ticket_data = cls.validate_pretix_ticket(conference, user, token)
        assert pretix_ident is not None
        assert ticket_data is not None

        # try to note down the ticket
        try:
            ticket = cls(conference=conference, user=user, ident=pretix_ident, data=ticket_data)
            ticket.save()
            logger.info('Pretix ident "%s" accepted.', pretix_ident)

        except IntegrityError:
            logger.error('User has a ticket already: %s', user)
            raise TicketValidationError(_('ConferenceMemberTicket__user_has_ticket_already'))

        # all went well, return the ticket
        return ticket
