import json
from datetime import datetime, timedelta, timezone
from pathlib import Path

from django.test import TestCase, override_settings

from core.models import Event, Room
from core.models.assemblies import Assembly
from core.models.conference import Conference, ConferenceMember
from core.models.schedules import ScheduleSource, ScheduleSourceImport, ScheduleSourceMapping
from core.models.users import PlatformUser
from core.schedules.base import BaseScheduleSupport, ScheduleTypeManager, filter_additional_data, schedule_time_to_timedelta
from core.schedules.schedulejson import ScheduleJSONSupport
from core.schedules.schedulejsonfrab import ScheduleJSONFrabSupport
from core.schedules.schedulexml import ScheduleXMLSupport


class FileBasedScheduleSupport(BaseScheduleSupport):
    identifier = 'test'
    readonly = True

    configuration_fields = {}

    @property
    def filename(self):
        url = str(self.schedule_source.import_url)
        assert url.startswith('file://')
        fn = url[len('file://') :]

        base_path = Path(__file__).parent
        fn = base_path.joinpath(fn).resolve()
        assert str(fn).startswith(f'{base_path}/'), 'Directory Traversal Attempt in Unit Tests? wow, so much shady'
        return fn

    def ready(self):
        return self.filename.exists()

    def fetch(self):
        with self.filename.open('r') as f:
            data = json.load(f)
        return data

    def __str__(self):
        return f'TEST{{{self.filename}}}'


class ScheduleTests(TestCase):
    def setUp(self):
        self.conference = Conference(
            slug='foo',
            name='Foo Conference',
            start=datetime(2021, 12, 1, 9, 0, 0, tzinfo=timezone.utc),
            end=datetime(2021, 12, 31, 15, 00, 00, tzinfo=timezone.utc),
        )
        self.conference.save()

        self.user = PlatformUser(username='bernd')
        self.user.save()
        ConferenceMember(conference=self.conference, user=self.user).save()

        self.assembly = Assembly(conference=self.conference, slug='fnord', name='Fnord Assembly', is_official=True)
        self.assembly.save()

        ScheduleTypeManager.register_class(FileBasedScheduleSupport.identifier, FileBasedScheduleSupport)
        ScheduleTypeManager.register_class(ScheduleJSONSupport.identifier, ScheduleJSONSupport)
        ScheduleTypeManager.register_class(ScheduleJSONFrabSupport.identifier, ScheduleJSONFrabSupport)
        ScheduleTypeManager.register_class(ScheduleXMLSupport.identifier, ScheduleXMLSupport)

    def test_utilitys(self):
        x = {'guid': 'foo', 'speakers': ['asdf', 'qwer'], 'title': 'unittest'}
        actual = filter_additional_data(x)
        self.assertNotIn('guid', actual)
        self.assertNotIn('title', actual)
        self.assertIn('speakers', actual)

        self.assertEqual(timedelta(minutes=45), schedule_time_to_timedelta('45'))
        self.assertEqual(timedelta(hours=1, minutes=30), schedule_time_to_timedelta('90'))
        self.assertEqual(timedelta(hours=1, minutes=42), schedule_time_to_timedelta('1:42'))

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_basic(self):
        src = ScheduleSource.objects.create(
            assembly=self.assembly,
            conference=self.conference,
            import_type=FileBasedScheduleSupport.identifier,
            import_url='file://import_data/basic.json',
        )
        self.assertFalse(src.has_running_import)

        job = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        self.assertFalse(src.has_running_import)

        # sanity checks pre-import
        self.assertEqual(0, self.assembly.rooms.count())
        self.assertEqual(0, self.assembly.events.count())

        # do the import
        job.do_import()
        self.assertFalse(src.has_running_import)

        # sanity checks post-import
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job.state)
        self.assertIsNotNone(job.data)
        self.assertEqual(1, self.assembly.rooms.count())
        self.assertEqual(1, self.assembly.events.count())
        self.assertIn('_activity', job.data)
        job_activity = job.data['_activity']
        self.assertEqual(2, len(job_activity))
        self.assertTrue(all(a['action'] == 'added' for a in job.data['_activity']))

        # check that room and event have been created
        r1 = self.assembly.rooms.get(name='Saal 1')
        e1 = self.assembly.events.get(slug='minkorrekt')
        self.assertIsNotNone(r1)
        self.assertIsNotNone(e1)
        self.assertEqual('Methodisch Inkorrekt', e1.name)

        # check that mappings exist
        self.assertEqual(2, src.mappings.count())

        r1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.ROOM, source_id='eins')
        self.assertEqual(r1_m.local_id, r1.id, "room's mapping's local_id doesn't match room id")

        e1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.EVENT, source_id='cej9dwoi')
        self.assertEqual(e1_m.local_id, e1.id, "event's mapping's local_id doesn't match event id")

        # check other attributes
        self.assertEqual(e1.room, r1)
        self.assertEqual(timedelta(minutes=90), e1.schedule_duration)

        # change the event's name and flag it to be skipped
        e1.name = 'Fnord'
        e1.save()
        e1_m.skip = True
        e1_m.save()

        # add a new event locally which was not imported
        e2 = self.assembly.events.create(conference=self.conference, slug='fnordnewsshow', name='Fnord News Show', room=r1, is_public=True)

        # redo the import
        job2 = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        job2.do_import()
        job2_activity = job2.data.get('_activity')
        self.assertIsNotNone(job2_activity)
        self.assertEqual(2, len(job2_activity))

        # check that the event has not been changed
        e1.refresh_from_db()
        self.assertEqual('Fnord', e1.name, 'name of skipped event has been changed')

        # check that the event was flagged
        job2_e1_activity_items = [x for x in job2_activity if x['type'] == 'event' and x['source_id'] == e1_m.source_id]
        self.assertEqual(1, len(job2_e1_activity_items))
        self.assertEqual('skipped', job2_e1_activity_items[0]['action'])

        # fake a previous mapping for the local event (which is not present in the import data)
        src.mappings.create(mapping_type=ScheduleSourceMapping.MappingType.EVENT, local_id=e2.pk, source_id='foobar')

        # redo the import
        job3 = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        job3.do_import()
        job3_activity = job3.data.get('_activity')
        self.assertEqual(3, len(job3_activity))

        # check that the local (and fake-mapped) event was marked 'missing'
        job3_e2_activity_items = [x for x in job3_activity if x['type'] == 'event' and x['local_id'] == str(e2.id)]
        self.assertEqual(1, len(job3_e2_activity_items))
        self.assertEqual('missing', job3_e2_activity_items[0]['action'])

        # redo the import with missing_event='recall'
        self.assertTrue(e2.is_public)
        src.import_configuration = {'missing_events': 'recall'}
        src.save()
        job4 = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        job4.do_import()
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job4.state)
        job4_activity = job4.data.get('_activity')

        # check that the local (and fake-mapped) event was marked 'missing'
        job4_e2_activity_items = [x for x in job4_activity if x['type'] == 'event' and x['local_id'] == str(e2.id)]
        self.assertEqual(1, len(job4_e2_activity_items))
        self.assertEqual('missing', job4_e2_activity_items[0]['action'])

        # but now the event should have been de-published
        e2.refresh_from_db()
        self.assertFalse(e2.is_public)

        # redo the import with missing_event='delete'
        src.import_configuration = {'missing_events': 'delete'}
        src.save()
        job5 = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        job5.do_import()
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job5.state)
        job5_activity = job5.data.get('_activity')

        # check that the local (and fake-mapped) event was marked 'missing'
        job5_e2_activity_items = [x for x in job5_activity if x['type'] == 'event' and x['local_id'] == str(e2.id)]
        self.assertEqual(1, len(job5_e2_activity_items))
        self.assertEqual('deleted', job5_e2_activity_items[0]['action'])

        # but now the event should have been de-published
        with self.assertRaises(Event.DoesNotExist):
            e2.refresh_from_db()

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_wildcard(self):
        r1 = self.assembly.rooms.create(
            conference=self.conference,
            id='f19b7ae0-5f2d-420c-ae4b-48e56a09ff15',
            room_type=Room.RoomType.LECTURE_HALL,
        )
        a2 = Assembly.objects.create(
            conference=self.conference,
            name='CCC',
            is_official=True,
            state=Assembly.State.ACCEPTED,
        )
        r2 = a2.rooms.create(
            conference=self.conference,
            id='4f9c1df8-926a-49e6-aa64-515e63846c12',
            room_type=Room.RoomType.WORKSHOP,
        )
        src = ScheduleSource.objects.create(
            assembly=None,
            conference=self.conference,
            import_type=FileBasedScheduleSupport.identifier,
            import_url='file://import_data/wildcard.json',
        )
        self.assertFalse(src.has_running_import)

        job = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        self.assertFalse(src.has_running_import)

        # sanity checks pre-import
        self.assertEqual(1, self.assembly.rooms.count())
        self.assertEqual(1, a2.rooms.count())
        self.assertEqual(0, self.assembly.events.count())
        self.assertEqual(0, a2.events.count())

        # do the import
        job.do_import()
        self.assertFalse(src.has_running_import)

        # sanity checks post-import: no new rooms but two events in different assemblies
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job.state)
        self.assertIsNotNone(job.data)
        self.assertEqual(1, self.assembly.rooms.count())
        self.assertEqual(1, a2.rooms.count())
        self.assertEqual(1, self.assembly.events.count())
        self.assertEqual(1, a2.events.count())
        self.assertIn('_activity', job.data)
        job_activity = job.data['_activity']
        self.assertEqual(4, len(job_activity))
        self.assertTrue(all(a['action'] == 'added' for a in job.data['_activity']))

        self.assertEqual(r1.id, str(Event.objects.get(slug='minkorrekt').room_id))
        self.assertEqual(r2.id, str(Event.objects.get(slug='blubb').room_id))

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_xml(self):
        src = ScheduleSource.objects.create(
            assembly=self.assembly,
            conference=self.conference,
            import_type=ScheduleXMLSupport.identifier,
            import_url=f"file://{Path(__file__).parent.joinpath('import_data', 'schedule-2021.xml')}",
        )
        self.assertFalse(src.has_running_import)

        job = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        self.assertFalse(src.has_running_import)

        # sanity checks pre-import
        self.assertEqual(0, self.assembly.rooms.count())
        self.assertEqual(0, self.assembly.events.count())

        # do the import
        job.do_import()
        self.assertFalse(src.has_running_import)

        # sanity checks post-import
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job.state)
        self.assertIsNotNone(job.data)
        self.assertEqual(1, self.assembly.rooms.count())
        self.assertEqual(1, self.assembly.events.count())
        self.assertIn('_activity', job.data)
        self.assertTrue(all(a['action'] == 'added' for a in job.data['_activity']))

        # check that room and event have been created
        r1 = self.assembly.rooms.get(name='Blue Room')
        e1 = self.assembly.events.get(slug='netzpolitik_in_der_schweiz')
        self.assertIsNotNone(r1)
        self.assertIsNotNone(e1)

        # check that mappings exist
        self.assertEqual(2, src.mappings.count())  # 2 = 1 room + 1 event
        r1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.ROOM, source_id='Blue Room', local_id=r1.id)
        e1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.EVENT, source_id='8264', local_id=e1.id)
        self.assertIsNotNone(r1_m)
        self.assertIsNotNone(e1_m)

        # check other attributes
        self.assertEqual(e1.room, r1)
        self.assertEqual(timedelta(minutes=60), e1.schedule_duration)

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_json2020(self):
        src = ScheduleSource.objects.create(
            assembly=self.assembly,
            conference=self.conference,
            import_type=ScheduleJSONSupport.identifier,
            import_url=f"file://{Path(__file__).parent.joinpath('import_data', 'schedule-2020.json')}",
        )
        self.check_json(src, rooms=1, events=1, speakers=2, room_name='Yellow Room', event_id='12345')

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_json2021(self):
        # Catch error for invalid user avatar url, we don't care about the message
        with self.assertLogs('core.models.users', level='ERROR'):
            src = ScheduleSource.objects.create(
                assembly=self.assembly,
                conference=self.conference,
                import_type=ScheduleJSONSupport.identifier,
                import_url=f"file://{Path(__file__).parent.joinpath('import_data', 'schedule-2021.json')}",
            )
            self.check_json(src, rooms=2, events=1, speakers=2, room_name='Gray Room', event_id='12345')

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_json_frab(self):
        self.conference = Conference(
            slug='37c3',
            name='37th Chaos Communication Congress',
            start=datetime(2023, 12, 27, 9, 0, 0, tzinfo=timezone.utc),
            end=datetime(2023, 12, 31, 15, 00, 0, tzinfo=timezone.utc),
        )
        self.conference.save()
        self.assembly = Assembly(name='CCC', slug='ccc', conference=self.conference, state=Assembly.State.PLACED)
        self.assembly.save()

        src = ScheduleSource.objects.create(
            assembly=self.assembly,
            conference=self.conference,
            import_type=ScheduleJSONFrabSupport.identifier,
            import_url=f"file://{Path(__file__).parent.joinpath('import_data', 'frab.schedule.json')}",
        )
        self.check_json(src, rooms=3, events=1, speakers=3, room_name='Saal 1', event_id='12338')

    def check_json(self, src: ScheduleSource, rooms: int, events: int, speakers: int, room_name: str, event_id: str):
        self.assertFalse(src.has_running_import)

        job = src.imports.create(state=ScheduleSourceImport.State.PREPARED)
        self.assertFalse(src.has_running_import)

        # sanity checks pre-import
        self.assertEqual(0, self.assembly.rooms.count())
        self.assertEqual(0, self.assembly.events.count())

        # do the import
        job.do_import()
        self.assertFalse(src.has_running_import)

        # sanity checks post-import
        self.assertEqual(ScheduleSourceImport.State.COMPLETED, job.state)
        self.assertIsNotNone(job.data)
        self.assertEqual(rooms, self.assembly.rooms.count())
        self.assertEqual(events, self.assembly.events.count())
        self.assertIn('_activity', job.data)
        self.assertTrue(all(a['action'] == 'added' for a in job.data['_activity']))

        # check that room and event have been created
        r1 = self.assembly.rooms.get(name=room_name)
        e1 = self.assembly.events.get(slug='opening')
        self.assertIsNotNone(r1)
        self.assertIsNotNone(e1)

        # check that mappings exist
        self.assertEqual(rooms + events + speakers, src.mappings.count())
        r1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.ROOM, source_id=room_name, local_id=r1.id)
        e1_m = src.mappings.get(mapping_type=ScheduleSourceMapping.MappingType.EVENT, source_id=event_id, local_id=e1.id)
        self.assertIsNotNone(r1_m)
        self.assertIsNotNone(e1_m)

        # check other attributes
        self.assertEqual(e1.room, r1)
        self.assertEqual(timedelta(minutes=30), e1.schedule_duration)
