import environ

from django.urls import reverse_lazy

from hub.settings.base import *  # noqa: F403 # pylint: disable=wildcard-import,unused-wildcard-import

default_env = environ.FileAwareEnv(
    # set casting, default value
    DJANGO_HOST=(str, ''),
    SERVE_ADMIN=(bool, False),
    SERVE_API=(bool, True),
    SERVE_BACKOFFICE=(bool, True),
    SERVE_FRONTEND=(bool, True),
    SERVE_MODULE_OUTPUT=(bool, True),
    SILENT_START=(bool, False),
    SECRET_KEY=(str, None),
    SSO_WARNING=(bool, True),
    DJANGO_SECRET_KEY=(str, None),
)


def print_banner(message: str, filler: str = '*'):
    if default_env('SILENT_START'):
        return
    print(filler * 72)
    print(filler, message)
    print(filler * 72)


# configure database using environment variable "DATABASE_URL"
DATABASES['default'] = default_env.db(default='postgis://localhost/hub')  # noqa: F405

# ensure the DJANGO_HOST variable's content is an allowed host (used for health check)
if (django_host := default_env('DJANGO_HOST')) is not None:
    if django_host not in ALLOWED_HOSTS:  # noqa:F405
        ALLOWED_HOSTS.append(django_host)  # noqa:F405

# set flags
SERVE_MODULE_OUTPUT = default_env('SERVE_MODULE_OUTPUT')
IS_ADMIN = default_env('SERVE_ADMIN')
if SERVE_MODULE_OUTPUT and IS_ADMIN:
    print_banner('Admin module has been selected and will be served', '-')
IS_API = default_env('SERVE_API')
if SERVE_MODULE_OUTPUT and IS_API:
    print_banner('API module has been selected and will be served', '-')
IS_BACKOFFICE = default_env('SERVE_BACKOFFICE')
if SERVE_MODULE_OUTPUT and IS_BACKOFFICE:
    print_banner('Backoffice module has been selected and will be served', '-')
IS_FRONTEND = default_env('SERVE_FRONTEND')
if SERVE_MODULE_OUTPUT and IS_FRONTEND:
    print_banner('Front-end module has been selected and will be served', '-')

# set SECRET_KEY from environment _before_ loading local_settings so that
# that can take precedence in case another key is specified there
assert SECRET_KEY is None, 'SECRET_KEY has already been set before default settings apply, something is wrong!'  # noqa: F405
SECRET_KEY = default_env('DJANGO_SECRET_KEY')  # noqa:F405
assert SECRET_KEY is None or not SECRET_KEY.startswith('/'), (
    'Your SECRET_KEY starts with a slash.'
    'If you intend to provide a filename use DJANGO_SECRET_KEY_FILE environment variable, otherwise choose another secret key.'
)

# Local Settings
try:
    from hub.local_settings import *  # noqa: F403,F401 # pylint: disable=wildcard-import,unused-wildcard-import
except ImportError:
    try:
        from hub.settings.local_settings import *  # noqa: F403,F401 # pylint: disable=wildcard-import,unused-wildcard-import
    except ImportError:
        print('Unable to load (optional) local_settings.py.')

# read SECRET_KEY from a file not stored in git
if DEBUG and SECRET_KEY is None:  # noqa: F405
    SECRET_FILE = BASE_DIR.joinpath('hub', '.settings.secret')  # noqa: F405
    try:
        with SECRET_FILE.open(encoding='UTF-8') as secret_file:
            SECRET_KEY = secret_file.read().strip()
    except OSError:
        try:
            SECRET_KEY = gen_secret_key()  # noqa: F405
            with SECRET_FILE.open('w', encoding='UTF-8') as secret:
                secret.write(SECRET_KEY)
                print('*' * 72)
                print('*', 'Written SECRET_KEY file:', SECRET_FILE)
                print('*' * 72)
        except OSError:
            raise Exception(f'Please create a {SECRET_FILE} file with random characters to serve as your Django SECRET_KEY!')

if DEBUG:  # noqa: F405
    INSTALLED_APPS.insert(INSTALLED_APPS.index('django.contrib.staticfiles') + 1, ('django.contrib.admindocs'))  # noqa: F405

# include API django api
if IS_API:
    # add 'api' app and ensure 'core' and it are before 'oauth2_provider' and 'rest_framework' for proper template loading
    INSTALLED_APPS.remove('core')  # noqa: F405
    INSTALLED_APPS.remove('rest_framework')  # noqa: F405
    INSTALLED_APPS.remove('oauth2_provider')  # noqa: F405
    INSTALLED_APPS.extend(['core', 'api', 'oauth2_provider', 'rest_framework'])  # noqa: F405

    # with API enabled default to use the API login screen
    LOGIN_URL = reverse_lazy('api:login')

# include backoffice project
if IS_BACKOFFICE:
    INSTALLED_APPS += ['backoffice', 'django_bootstrap5', 'widget_tweaks']  # noqa: F405

    # with backoffice enabled default to use its login screen
    LOGIN_URL = reverse_lazy('backoffice:login')  # this overwrites previous LOGIN_URL assignment(s) on purpose

# activate oauth2 toolkit in both API and backoffice
if IS_API or IS_BACKOFFICE:
    INSTALLED_APPS += ['corsheaders']  # noqa: F405
    # insert CORS middleware before Django's CommonMiddleware
    MIDDLEWARE.insert(0, 'corsheaders.middleware.CorsMiddleware')  # noqa: F405

    # CORS spec
    CORS_ALLOWED_ORIGINS = [
        # TODO: make this an env var
    ]
    CORS_ALLOWED_ORIGIN_REGEXES = [
        # TODO: make this an env var
    ]
    CORS_URLS_REGEX = r'^/api/.*$'

    if DEBUG:  # noqa: F405
        CORS_ALLOWED_ORIGIN_REGEXES.append(r'^https://django-oauth-toolkit\.herokuapp\.com/consumer/')
        OAUTH2_PROVIDER.get('ALLOWED_REDIRECT_URI_SCHEMES', []).append('http')  # noqa: F405

# include frontend (plainui) project
if IS_FRONTEND:
    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.jinja2.Jinja2',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                    'csp.context_processors.nonce',
                ],
                'environment': 'plainui.jinja2.environment',
            },
        },
        *TEMPLATES,  # noqa: F405
    ]

    # TODO: make up a proper redirect mechanism
    LOGIN_URL = reverse_lazy('plainui:login')  # this overwrites previous LOGIN_URL assignment(s) on purpose

# activate sentry upon request
if SENTRY_ENDPOINT is not None:  # noqa: F405
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration

    release = APP_VERSION_INFO.get('tag', APP_VERSION_INFO.get('commit', None))  # noqa: F405

    sentry_sdk.init(
        dsn=SENTRY_ENDPOINT,  # noqa: F405
        integrations=[DjangoIntegration()],
        environment=SENTRY_ENVIRONMENT,  # noqa: F405
        # trace sample rate in percent
        traces_sample_rate=SENTRY_TRACES_SAMPLE_RATE,  # noqa: F405
        # profiling rate (percentage of traces)
        profiles_sample_rate=SENTRY_PROFILES_SAMPLE_RATE,  # noqa: F405
        # do not send user information (django.contrib.auth)
        send_default_pii=False,
        release=release,
    )

if IS_API and SSO_SECRET is None:  # noqa: F405
    SSO_SECRET = SECRET_KEY[::-1]
    if default_env('SSO_WARNING'):
        print_banner('Generated a SSO_SECRET from your SECRET_KEY, this is probably not what you want!')

if '*' in METRICS_SERVER_IPS:  # noqa: F405
    print_banner('Warning: /metrics/ is accessible from every IP address ("*"). Set METRICS_SERVER_IPS="<prometheus server ip>".')

TEST_RUNNER = 'core.tests.runner.PostgresSchemaTestRunner'
