"""hub URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from typing import Any

from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path

from hub import views

services = {
    'admin': settings.IS_ADMIN,
    'api': settings.IS_API,
    'backoffice': settings.IS_BACKOFFICE,
    'frontend': settings.IS_FRONTEND,
}

handler403 = views.handle403  # noqa: F811


def is_only(service: str) -> bool:
    return not any(v for k, v in services.items() if k != service)


def add_service(service: str, prefix: str, service_patterns: Any, path_name: str = '{prefix}', always_prefix: bool = False):
    if not services[service]:
        return []
    return [
        path(
            prefix if always_prefix else path_name.format(prefix='' if is_only(service) else prefix),
            service_patterns,
        )
    ]


urlpatterns = [
    path('.well-known/health', views.health),
    path('.well-known/version', views.version),
]


# If no services are selected raise a 500 on index
if not any(services.values()):
    urlpatterns += [path('', views.index)]

# Go through the services and add them to the patterns if it's selected
# Any service that is not the last one is served with a prefix.
# The last service is served at root.
# Priority: Frontend -> Backoffice -> API -> Admin
if 'django.contrib.staticfiles' in settings.INSTALLED_APPS:
    # Add the Admin Docs package to show current Documentation generated from docstrings
    urlpatterns += add_service('admin', 'c3admin/docs/', include('django.contrib.admindocs.urls'))
urlpatterns += add_service('admin', 'c3admin/', admin.site.urls)
urlpatterns += (path('i18n/setlang/', views.set_language, name='set_language'),)
services.pop('admin')

# API
urlpatterns += add_service(
    'api',
    'api/',
    service_patterns=include('api.urls'),
)
urlpatterns += add_service('api', 'api/', path_name='{prefix}auth/', service_patterns=include('rest_framework.urls', namespace='rest_framework'))
urlpatterns += add_service('api', 'metrics/', include('api.urls_metrics', namespace='metrics'), always_prefix=True)
services.pop('api')

# Backoffice
urlpatterns += add_service('backoffice', 'backoffice/', include('backoffice.urls', namespace='backoffice'))
services.pop('backoffice')

# SSO (Frontend)
urlpatterns += add_service('frontend', 'sso/', include('core.urls_sso', namespace='oauth2'), always_prefix=True)

# PlainUI (Frontend)
if services['frontend']:
    urlpatterns += i18n_patterns(
        path('', include('plainui.urls', namespace='plainui')),
        prefix_default_language=True,
    )

if settings.DEBUG:
    # serve media (uploaded files) with development server
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    # add path for debug toolbar if it is present
    if 'debug_toolbar' in settings.INSTALLED_APPS:
        try:
            import debug_toolbar

            urlpatterns += [
                path('__debug__/', include(debug_toolbar.urls, namespace='djdt')),
            ]
        except ImportError:
            pass

# serve media files if specifically requested to do so
if settings.FORCE_SERVE_MEDIA:
    urlpatterns += static('/media/', document_root=settings.MEDIA_ROOT)

# debug endpoint: reply with request headers
if settings.HEADERS_DEBUG_ENDPOINT_TOKEN is not None:
    urlpatterns += [
        path('.well-known/headers', views.debug_headers),
    ]
