import re
from smtplib import SMTPException
from unittest.mock import patch

from django.contrib.auth import SESSION_KEY, get_user
from django.contrib.messages import ERROR, get_messages
from django.core import mail
from django.core.exceptions import NON_FIELD_ERRORS
from django.http import SimpleCookie
from django.test import override_settings
from django.urls import reverse
from django.utils.translation import override as override_locale

from core.models import (
    PlatformUser,
    UserCommunicationChannel,
)
from core.tests.mock import mocktrans

from plainui.tests.test_views import ViewsTestBase


class AuthViewTest(ViewsTestBase):
    @override_locale('en')
    def test_password_change_get(self):
        self.assertNeedsLogin(reverse('plainui:password_change'), check_conference_member=False, check_user=True)
        resp = self.client.get(reverse('plainui:password_change'))
        self.assertEqual(resp.context_data['conf'], self.conf)

    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    def test_password_change_post(self):
        self.user.set_password('old_password')
        self.user.save()
        self.client.force_login(self.user)  # got logged out by set_password

        self.assertNeedsLogin(reverse('plainui:password_change'), check_conference_member=False, check_user=True)
        resp = self.client.post(
            reverse('plainui:password_change'),
            {
                'old_password': 'invalid',
                'new_password1': 'new_password',
                'new_password2': 'new_password',
            },
        )
        self.assertTrue(resp.context_data['form']['old_password'].errors)

        self.assertNeedsLogin(reverse('plainui:password_change'), check_conference_member=False, check_user=True)
        resp = self.client.post(
            reverse('plainui:password_change'),
            {
                'old_password': 'old_password',
                'new_password1': 'new_password',
                'new_password2': 'new_password',
            },
        )
        self.assertRedirects(resp, reverse('plainui:userprofile'))
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('new_password'))
        self.assertSetsMessage(resp, 'Password changed successfully.')

    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    def test_password_reset(self):
        self.user.set_password('forgotten')
        self.user.save()
        channel = UserCommunicationChannel(user=self.user, channel=UserCommunicationChannel.Channel.MAIL, is_verified=True, address='no@where.test')
        channel.save()

        resp = self.client.get(reverse('plainui:password_reset'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(len(mail.outbox), 0)
        self.assertNotContains(resp, 'Reset password link invalid')

        resp = self.client.post(reverse('plainui:password_reset'), {'email': 'doesnot@exist.test'})
        self.assertEqual(len(mail.outbox), 0)

        resp = self.client.post(reverse('plainui:password_reset'), {'email': 'no@where.test'})
        self.assertRedirects(resp, reverse('plainui:password_reset_done'))
        self.assertEqual(len(mail.outbox), 1)
        sent_mail = mail.outbox[0]
        self.assertEqual(sent_mail.to, ['no@where.test'])
        self.assertEqual(sent_mail.subject, 'Password reset on testserver')
        reset_url_pattern = reverse('plainui:password_reset_confirm', kwargs={'uidb64': 'uidb64', 'token': 'token'})
        reset_url_pattern = re.escape(reset_url_pattern)
        reset_url_pattern = reset_url_pattern.replace('uidb64', '([^/]+)')
        reset_url_pattern = reset_url_pattern.replace('token', '([^/]+)')
        m = re.search(reset_url_pattern, sent_mail.body)
        self.assertIsNotNone(m, 'password reset confirm url is missing in sent mail!')

    @override_locale('en')
    def test_password_reset_done(self):
        resp = self.client.get(reverse('plainui:password_reset_done'))
        self.assertEqual(resp.context_data['conf'], self.conf)

    @override_settings(LANGUAGE_CODE='en', AUTH_PASSWORD_VALIDATORS=[])
    @override_locale('en')
    def test_password_reset_confirm(self):
        from datetime import timedelta

        from django.contrib.auth.tokens import default_token_generator as token_generator
        from django.contrib.auth.views import INTERNAL_RESET_SESSION_TOKEN
        from django.utils.encoding import force_bytes
        from django.utils.http import urlsafe_base64_encode

        self.user.set_password('forgotten')
        self.user.save()
        self.client.cookies = SimpleCookie()  # password reset only works when not logged in
        uidb = urlsafe_base64_encode(force_bytes(self.user.pk))
        token = token_generator.make_token(self.user)

        resp = self.client.get(reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': token}))
        self.assertEqual(self.client.session[INTERNAL_RESET_SESSION_TOKEN], token)
        self.assertRedirects(resp, reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': 'set-password'}))

        resp = self.client.get(reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': token}))
        self.assertEqual(self.client.session[INTERNAL_RESET_SESSION_TOKEN], token)
        self.assertRedirects(resp, reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': 'set-password'}))

        resp = self.client.get(reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': 'set-password'}))
        self.assertEqual(resp.context_data['conf'], self.conf)

        resp = self.client.post(
            reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': 'set-password'}),
            {
                'new_password1': '4',  # chosen by fair dice roll
                'new_password2': '4',  # guranteed to be random
            },
        )
        self.assertRedirects(resp, reverse('plainui:password_reset_complete'))

        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('4'))

        expired_token = token_generator._make_token_with_timestamp(
            self.user, token_generator._num_seconds(token_generator._now() - timedelta(days=365)), secret=None
        )
        expired_link = reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': expired_token})
        expired_resp = self.client.get(expired_link, follow=True)
        self.assertRedirects(expired_resp, f"{reverse('plainui:password_reset')}?retry=True")
        self.assertContains(expired_resp, 'Reset password link invalid')

        invalid_session_resp = self.client.get(reverse('plainui:password_reset_confirm', kwargs={'uidb64': uidb, 'token': 'set-password'}))
        self.assertNotIn(INTERNAL_RESET_SESSION_TOKEN, self.client.session)
        self.assertRedirects(invalid_session_resp, f"{reverse('plainui:password_reset')}?retry=True")

    @override_locale('en')
    def test_password_reset_complete(self):
        resp = self.client.get(reverse('plainui:password_reset_done'))
        self.assertEqual(resp.context_data['conf'], self.conf)

    @override_locale('en')
    def test_logout(self):
        self.assertIn(SESSION_KEY, self.client.session)
        self.assertTrue(get_user(self.client).is_authenticated)
        self.assertNotEqual('', self.client.session[SESSION_KEY])

        resp = self.client.post(reverse('plainui:logout'))
        self.assertFalse(get_user(self.client).is_authenticated)
        self.assertNotIn(SESSION_KEY, self.client.session)
        self.assertRedirects(resp, reverse('plainui:login'))


class LoginViewTests(ViewsTestBase):
    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    def test_login(self):
        self.user.set_password('asdf')
        self.user.theme = PlatformUser.Theme.DARK
        self.user.save()

        resp = self.client.get(reverse('plainui:login'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['next'], '')
        self.assertFalse(resp.context_data['user'].is_authenticated)

        resp = self.client.get(reverse('plainui:login'), {'next': 'http://some.where/foo'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['next'], '')
        self.assertFalse(resp.context_data['user'].is_authenticated)

        resp = self.client.get(reverse('plainui:login'), {'next': reverse('plainui:index')})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['next'], reverse('plainui:index'))
        self.assertFalse(resp.context_data['user'].is_authenticated)

        resp = self.client.post(
            reverse('plainui:login'),
            {
                'username': 'testuser',
                'password': 'invalid',
            },
        )
        self.assertIsNone(self.client.session.get('SESSION_KEY'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['next'], '')
        self.assertEqual(resp.context_data['form']['username'].value(), 'testuser')
        self.assertTrue(resp.context_data['form'].errors)
        self.assertFalse(resp.context_data['user'].is_authenticated)

        resp = self.client.post(
            reverse('plainui:login'),
            {
                'username': 'invalid',
                'password': 'asdf',
            },
        )
        self.assertNotIn(SESSION_KEY, self.client.session)
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['next'], '')
        self.assertEqual(resp.context_data['form']['username'].value(), 'invalid')
        self.assertTrue(resp.context_data['form'].errors)
        self.assertFalse(resp.context_data['user'].is_authenticated)

        resp = self.client.post(
            reverse('plainui:login'),
            {
                'username': 'testuser',
                'password': 'asdf',
            },
        )
        self.assertIn(SESSION_KEY, self.client.session)
        self.assertEqual(self.client.session['theme'], self.user.theme)
        self.assertRedirects(resp, reverse('plainui:index'))

        self.client.cookies = SimpleCookie()
        resp = self.client.post(
            reverse('plainui:login'),
            {
                'username': 'testuser',
                'password': 'asdf',
                'distractfree_theme': 'on',
            },
        )
        self.assertIn(SESSION_KEY, self.client.session)
        self.assertEqual(self.client.session['theme'], PlatformUser.Theme.DARK)
        self.assertRedirects(resp, reverse('plainui:index'))

        # we've now done 4 requests in short succession, will do a couple more dummy requests and then next request will run into the rate limiting
        for i in range(5):
            self.client.cookies = SimpleCookie()
            resp = self.client.post(
                reverse('plainui:login'),
                {
                    'username': 'testuser',
                    'password': 'asdf',
                },
            )
        self.assertTrue(resp.context_data['request'].limited)
        self.assertTrue(resp.context_data['form'].errors)
        self.assertFalse(resp.context_data['user'].is_authenticated)
        self.assertNotIn(SESSION_KEY, self.client.session)


class RegistrationTests(ViewsTestBase):
    @override_locale('en')
    @override_settings(RATELIMIT_ENABLE=False)
    def test_registration(self):
        resp = self.client.get(reverse('plainui:signup'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(len(mail.outbox), 0)

        resp = self.client.post(reverse('plainui:signup'), {'username': 'unittest', 'password1': 'mis', 'password2': 'match', 'email': 'doesnot@exist.test'})
        self.assertTrue(resp.context_data['form']['password2'].errors)
        self.assertEqual(len(mail.outbox), 0)

        resp = self.client.post(
            reverse('plainui:signup'),
            {
                'username': 'unittest',
                'password1': 'Fnord4711',
                'password2': 'Fnord4711',
                'email': 'unittest@example.com',
            },
        )
        self.assertRedirects(resp, reverse('plainui:signup_done'))
        self.assertEqual(len(mail.outbox), 1)
        sent_mail = mail.outbox[0]
        self.assertEqual(sent_mail.to, ['unittest@example.com'])
        self.assertEqual(sent_mail.subject, 'Activate your ' + self.conf.name + ' account')
        reset_url_pattern = reverse('plainui:signup_activate', kwargs={'uid_b64': 'uidb64', 'channel_id': 42, 'token': 'token-token'})
        reset_url_pattern = re.escape(reset_url_pattern)
        reset_url_pattern = reset_url_pattern.replace('uidb64', '([^/]+)')
        reset_url_pattern = reset_url_pattern.replace('42', r'(\d+)')
        reset_url_pattern = reset_url_pattern.replace('token\\-token', '([^/]+)')
        m = re.search(reset_url_pattern, sent_mail.body)
        self.assertIsNotNone(m, 'confirm url is missing in sent mail!')

        # now click the activation link and check that the user gets activated
        u = PlatformUser.objects.get(username='unittest')
        self.assertFalse(u.is_active)
        resp = self.client.get(m.group(0))
        self.assertRedirects(resp, reverse('plainui:index'))
        u.refresh_from_db()
        self.assertTrue(u.is_active)
        self.assertListEqual(list(u.conferences.all()), list(self.conf.users.filter(user__username='unittest')))

    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    def test_mail_failure(self):
        with patch('django.core.mail.EmailMultiAlternatives.send') as email_mock:
            email_mock.side_effect = SMTPException

            resp = self.client.get(reverse('plainui:signup'))
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(len(mail.outbox), 0)

            resp = self.client.post(
                reverse('plainui:signup'), {'username': 'unittest', 'password1': 'mis', 'password2': 'match', 'email': 'doesnot@exist.test'}
            )
            self.assertTrue(resp.context_data['form']['password2'].errors)
            self.assertEqual(len(mail.outbox), 0)

            with patch('core.views.auth._', side_effect=mocktrans):
                resp = self.client.post(
                    reverse('plainui:signup'),
                    {
                        'username': 'unittest',
                        'password1': 'Fnord4711',
                        'password2': 'Fnord4711',
                        'email': 'unittest@example.com',
                    },
                )
            messages = list(get_messages(resp.wsgi_request))
            self.assertEqual(len(messages), 1)
            message = messages[0]
            self.assertEqual(message.level, ERROR)
            self.assertEqual(message.message, 'Registration__email__sending__failure_translated')
            self.assertEqual(resp.status_code, 200)
            self.assertTrue(resp.context_data['form'].errors[NON_FIELD_ERRORS])
            self.assertEqual(len(mail.outbox), 0)
