import re
import uuid
from datetime import UTC, datetime, timedelta
from unittest.mock import patch
from urllib.parse import quote

import jwt
from freezegun import freeze_time

from django.contrib import messages
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseForbidden, SimpleCookie
from django.test import Client, RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import localize
from django.utils.timezone import localtime
from django.utils.translation import override as override_locale

from core.choices import UserBadgeVisibility
from core.markdown import render_markdown
from core.models import (
    Assembly,
    AssemblyMember,
    Badge,
    BadgeToken,
    BulletinBoardEntry,
    Conference,
    ConferenceMember,
    ConferenceMemberTicket,
    ConferenceTag,
    ConferenceTrack,
    DereferrerStats,
    DirectMessage,
    Event,
    EventParticipant,
    Lock,
    MarkdownMeta,
    PlatformUser,
    Project,
    Room,
    RoomLink,
    StaticPage,
    StaticPageRevision,
    TagItem,
    UserBadge,
    UserDereferrerAllowlist,
)
from core.templatetags.hub_absolute import hub_absolute
from core.tests.mock import mocktrans
from hub.http import HttpResponseRateLimited

from plainui.forms import ProfileEditForm
from plainui.views import ConferenceRequiredMixin, LandingView


# from https://github.com/Grollicus/unittest_patterns/blob/master/unittest_patterns/__init__.py
class Pattern:
    def __req__(self, lhs):
        return self.__eq__(lhs)

    __hash__ = None


# from https://github.com/Grollicus/unittest_patterns/blob/master/unittest_patterns/__init__.py
class Any(Pattern):
    """Equals everything"""

    def __eq__(self, rhs):
        return True


class ThingWithLength(Pattern):
    def __init__(self, _len):
        self.len = _len

    def __eq__(self, rhs):
        return len(rhs) == self.len


TEST_CONF_ID = uuid.uuid4()
TEST_CONF_ID_2 = uuid.uuid4()
TEST_CONF_ID_3 = uuid.uuid4()


@override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class ViewsTestBase(TestCase):
    fixtures = [
        'bootstrap_auth_groups',
    ]

    def setUp(self):
        self.conf = Conference(
            id=TEST_CONF_ID,
            name='conf_asdf',
            slug='slug1',
            start=datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC),
            end=datetime(2020, 1, 3, 0, 0, 0, tzinfo=UTC),
            is_public=True,
            require_ticket=True,
            require_login=True,
        )
        self.conf.save()
        self.user = PlatformUser(username='testuser', email='no@where.test')
        self.user.save()
        self.conference_member = ConferenceMember(conference=self.conf, user=self.user, has_ticket=True)
        self.conference_member.save()
        self.user_no_conf = PlatformUser(username='testuser_no_conf', email='somewhere@else.test')
        self.user_no_conf.save()
        self.user_no_ticket = PlatformUser(username='testuser_no_ticket', email='somewhere@here.test')
        self.user_no_ticket.save()
        self.conference_member_no_ticket = ConferenceMember(conference=self.conf, user=self.user_no_ticket, has_ticket=False)
        self.conference_member_no_ticket.save()

        ConferenceRequiredMixin._conf = None
        self.client.force_login(self.user)

    def assertNeedsLogin(self, url, data=None, post=False, check_conference_member=True, check_user=False):
        """
        Checks that the given View requires a user login if the conference is configured to be non-public.

        If `check_conference_member` is `True` (the default), also checks that the user needs a valid ticket,
        if the conference is configured to require tickets.

        If `check_user` is True, also checks that the View requires a user even if the conference is configured to be public

        This is the test pendant to `plainui.views.ConferenceRequiredMixin`.
        """
        old_cookies = self.client.cookies
        try:
            self.client.cookies = SimpleCookie()
            ConferenceRequiredMixin._test_cork = True
            _request = self.client.post if post else self.client.get

            # check that view requires user login if configured non-public
            resp = _request(url, data=data, follow=False)
            self.assertRedirects(resp, reverse('plainui:login') + '?next=' + quote(url))

            self.conf.require_login = False
            self.conf.require_ticket = False
            self.conf.save()
            ConferenceRequiredMixin._conf = None
            if check_user:
                # check that the view requires a user login even if conference is configured public
                resp = _request(url, data=data, follow=False)
                self.assertRedirects(resp, reverse('plainui:login') + '?next=' + quote(url))
            else:
                # check that the view requires no user login if conference is configured public
                resp = _request(url, data=data, follow=False)
                self.assertEqual(resp.status_code, 200)

            self.conf.require_login = True
            self.conf.require_ticket = True
            self.conf.save()
            ConferenceRequiredMixin._conf = None

            if check_conference_member:
                # check that view needs a valid ticket
                self.client.force_login(self.user_no_conf)
                resp = _request(url, data=data, follow=False)
                self.assertRedirects(resp, reverse('plainui:redeem_token'))

                # .. and doesn't fail whether there is no ConferenceMember or no ticket
                self.client.force_login(self.user_no_ticket)
                resp = _request(url, data=data, follow=False)
                self.assertRedirects(resp, reverse('plainui:redeem_token'))

                # check that the view needs no valid ticket if conference is required not to require tickets
                self.conf.require_ticket = False
                self.conf.save()
                ConferenceRequiredMixin._conf = None
                self.client.force_login(self.user_no_conf)
                resp = _request(url, data=data, follow=False)
                self.assertEqual(resp.status_code, 200)

                # .. and doesn't fail even if there is a conferenceuser but no ticket
                self.client.force_login(self.user_no_ticket)
                resp = _request(url, data=data, follow=False)
                self.assertEqual(resp.status_code, 200)
        finally:
            self.client.cookies = old_cookies
            ConferenceRequiredMixin._test_cork = False

            self.conf.require_login = True
            self.conf.require_ticket = True
            self.conf.save()
            ConferenceRequiredMixin._conf = None

    def assertNeedsNoLogin(self, url, data=None, post=False):
        old_cookies = self.client.cookies
        try:
            self.client.cookies = SimpleCookie()
            ConferenceRequiredMixin._test_cork = True
            _request = self.client.post if post else self.client.get

            # check that view requires no login even for anonymous user
            resp = _request(url, data=data, follow=False)
            self.assertEqual(resp.status_code, 200)
        finally:
            self.client.cookies = old_cookies
            ConferenceRequiredMixin._test_cork = False

    def assertSetsMessage(self, request, msg: str, level=messages.SUCCESS):
        msgs = list(messages.get_messages(request.wsgi_request))
        self.assertEqual(len(msgs), 1, 'Expected exactly one Message')
        self.assertEqual(str(msgs[0].message), msg)
        self.assertEqual(msgs[0].level, level)


@override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class ViewsTest(ViewsTestBase):
    @override_locale('en')
    def test_EventView(self):
        assembly = Assembly(conference=self.conf, name='DUMMY', slug='dummy', state=Assembly.State.ACCEPTED, is_official=True)
        assembly.save()

        tag = ConferenceTag(conference=self.conf, slug='test_tag', is_public=True, value_type='simple')
        tag.save()
        tag_hidden = ConferenceTag(conference=self.conf, slug='hidden_tag', is_public=False, value_type='simple')
        tag_hidden.save()

        event = Event(
            conference=self.conf,
            slug='event1',
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
            kind=Event.Kind.OFFICIAL,
        )
        event.save()

        tagitem = TagItem(tag=tag, target_type=ContentType.objects.get_for_model(Event), target_id=event.pk)
        tagitem.save()
        tagitem_hidden = TagItem(tag=tag_hidden, target_type=ContentType.objects.get_for_model(Event), target_id=event.pk)
        tagitem_hidden.save()

        self.assertNeedsLogin(reverse('plainui:event', kwargs={'event_slug': event.slug}))
        resp = self.client.get(reverse('plainui:event', kwargs={'event_slug': event.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['event'], event)
        self.assertEqual(list(resp.context_data['tags']), [tagitem])

        event.kind = Event.Kind.ASSEMBLY
        event.save()
        resp = self.client.get(reverse('plainui:event', kwargs={'event_slug': event.slug}))

        event.favorite_of.add(self.user)
        session = self.client.session
        session['fav_e'] = None
        session.save()
        resp = self.client.get(reverse('plainui:event', kwargs={'event_slug': event.slug}))
        self.assertTrue(resp.context_data['is_favorite'])

    @override_locale('en')
    def test_TagView(self):
        assembly1 = Assembly(conference=self.conf, slug='d', name='DUMMY', state=Assembly.State.ACCEPTED)
        assembly1.save()
        assembly2 = Assembly(conference=self.conf, slug='2', name='2nd', state=Assembly.State.ACCEPTED)
        assembly2.save()

        tag = ConferenceTag(conference=self.conf, slug='test_tag', is_public=True, value_type='simple')
        tag.save()

        event1 = Event(
            conference=self.conf,
            assembly=assembly1,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event1.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly1,
            name='Event1_2',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()

        TagItem(tag=tag, target_type=ContentType.objects.get_for_model(Event), target_id=event1.pk).save()
        TagItem(tag=tag, target_type=ContentType.objects.get_for_model(Assembly), target_id=assembly2.pk).save()

        self.assertNeedsLogin(reverse('plainui:tag', kwargs={'tag_slug': tag.slug}))
        resp = self.client.get(reverse('plainui:tag', kwargs={'tag_slug': tag.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['tag'], tag)
        self.assertEqual(list(resp.context_data['events']), [event1])
        self.assertEqual(list(resp.context_data['assemblies']), [assembly2])

    @override_locale('en')
    def test_AssemblyView(self):
        speaker1 = PlatformUser(username='speaker1')
        speaker1.save()
        speaker2 = PlatformUser(username='speaker2')
        speaker2.save()
        member1 = PlatformUser(username='member1')
        member1.save()
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        AssemblyMember(member=speaker1, is_representative=True, can_manage_assembly=True, show_public=True, assembly=assembly).save()
        AssemblyMember(member=speaker2, is_representative=True, can_manage_assembly=True, show_public=False, assembly=assembly).save()
        AssemblyMember(member=member1, show_public=True, assembly=assembly).save()
        tag = ConferenceTag(conference=self.conf, slug='test_tag', is_public=True, value_type='simple')
        tag.save()
        tag_item = TagItem(tag=tag, target_type=ContentType.objects.get_for_model(Assembly), target_id=assembly.pk)
        tag_item.save()
        event1 = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event1.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_2',
            is_public=False,
            schedule_start=datetime(2020, 1, 2, 0, 0, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()
        project1 = Project(
            conference=self.conf,
            assembly=assembly,
            is_public=True,
            blocked=False,
            name='Project1',
        )
        project1.save()
        project2 = Project(
            conference=self.conf,
            assembly=assembly,
            is_public=False,
            blocked=False,
            name='Project1',
        )
        project2.save()
        room1 = Room(
            conference=self.conf,
            assembly=assembly,
            room_type=Room.RoomType.LECTURE_HALL,
            name='Roomq',
        )
        room1.save()

        self.assertNeedsLogin(reverse('plainui:assembly', kwargs={'assembly_slug': assembly.slug}))
        resp = self.client.get(reverse('plainui:assembly', kwargs={'assembly_slug': assembly.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['assembly'], assembly)
        self.assertFalse(resp.context_data['is_favorite'])
        context_assembly = resp.context_data['assembly']
        self.assertListEqual(list(context_assembly.public_tags), [tag_item])
        self.assertEqual(list(context_assembly.public_events), [event1])
        self.assertEqual(list(context_assembly.public_projects), [project1])
        self.assertEqual(list(context_assembly.public_rooms), [room1])
        self.assertEqual([sp.member for sp in context_assembly.public_members], [speaker1, member1])

        assembly.favorite_of.add(self.user)
        session = self.client.session
        session['fav_a'] = None
        session.save()
        resp = self.client.get(reverse('plainui:assembly', kwargs={'assembly_slug': assembly.slug}))
        self.assertTrue(resp.context_data['is_favorite'])

        sos_assembly = Assembly(conference=self.conf, slug='sos_assembly', name='SOS Assembly', state=Assembly.State.PLACED)
        sos_assembly.save()

        # redirects to the sos page when accessing the sos assembly as the existence of the assembly page caused some confusion
        self.conf.self_organized_sessions_assembly = sos_assembly
        self.conf.save()
        resp = self.client.get(reverse('plainui:assembly', kwargs={'assembly_slug': sos_assembly.slug}))
        self.assertRedirects(resp, reverse('plainui:sos'))

    @override_locale('en')
    def test_AssembliesView(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()

        self.assertNeedsLogin(reverse('plainui:assemblies'))
        resp = self.client.get(reverse('plainui:assemblies'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['assemblies']), [assembly])

    @override_locale('en')
    def test_AssembliesAllView(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()

        self.assertNeedsLogin(reverse('plainui:assemblies_all'))
        resp = self.client.get(reverse('plainui:assemblies_all'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['object_list']), [assembly])

    @override_locale('en')
    def test_AssembliesEventsView(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Some Room')
        room.save()
        room_public = Room(conference=self.conf, assembly=assembly, name='Public Fahrplan Room', is_public_fahrplan=True)
        room_public.save()

        event = Event(
            conference=self.conf,
            assembly=assembly,
            room=room,
            slug='Event1_1',
            name='Event1_1',
            is_public=True,
            kind=Event.Kind.ASSEMBLY,
            schedule_start=datetime(2020, 1, 2, 0, 45, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly,
            room=room,
            slug='Event1_2',
            name='Event1_2',
            is_public=True,
            kind=Event.Kind.OFFICIAL,
            schedule_start=datetime(2020, 1, 2, 0, 50, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()
        event3 = Event(
            conference=self.conf,
            assembly=assembly,
            room=room_public,
            slug='Event1_3',
            name='Event1_3',
            is_public=True,
            kind=Event.Kind.OFFICIAL,
            schedule_start=datetime(2020, 1, 2, 0, 55, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event3.save()
        event4 = Event(
            conference=self.conf,
            assembly=assembly,
            room=room,
            slug='Event1_4',
            name='Event1_4',
            is_public=False,
            kind=Event.Kind.OFFICIAL,
            schedule_start=datetime(2020, 1, 2, 1, 0, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event4.save()

        self.assertNeedsLogin(reverse('plainui:assemblies_events'))
        with freeze_time(datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC)):
            resp = self.client.get(reverse('plainui:assemblies_events'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['events_from_assemblies']), [event])
        self.assertEqual(list(resp.context_data['events_upcoming']), [])
        self.assertEqual(list(resp.context_data['is_favorite_events']), [])

        self.user.favorite_assemblies.add(assembly)
        self.user.favorite_events.add(event)
        session = self.client.session
        session['fav_e'] = session['fav_a'] = session['sch_e'] = None
        session.save()

        self.client.force_login(self.user)  # sonst funktioniert der favoriten cache reset nicht. dirty hack, irgendwann ausbauen ^.^
        with freeze_time(datetime(2020, 1, 2, 0, 30, 0, tzinfo=UTC)):
            resp = self.client.get(reverse('plainui:assemblies_events'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['events_from_assemblies']), [event])
        self.assertEqual(list(resp.context_data['events_upcoming']), [event, event2, event3])
        self.assertEqual(list(resp.context_data['is_favorite_events']), [str(event.pk)])

    @override_locale('en')
    def test_StaticPageView(self):
        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1')
        r1.save()
        r1.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title 2', page=sp_de, body='Body 2')
        r1_de.save()
        r1_de.set_public()

        sp_private_conf = StaticPage(conference=self.conf, slug='test1_private_conf', language='en', privacy=StaticPage.Privacy.CONFERENCE)
        sp_private_conf.save()
        r1_private_conf = StaticPageRevision(title='Title 1 Private Conf', page=sp_private_conf, body='Body 1 Private Conf')
        r1_private_conf.save()
        r1_private_conf.set_public()

        sp_private_perm = StaticPage(conference=self.conf, slug='test1_private_perm', language='en', privacy=StaticPage.Privacy.PERM)
        sp_private_perm.save()
        r1_private_perm = StaticPageRevision(title='Title 1 Private Perm', page=sp_private_perm, body='Body 1 Private Perm')
        r1_private_perm.save()
        r1_private_perm.set_public()

        self.assertNeedsLogin(reverse('plainui:static_page', kwargs={'page_slug': sp.slug}), check_conference_member=False)
        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'].title, 'Title 1')
        self.assertIn('Body 1', resp.context_data['page'].body_html)
        self.assertEqual(resp.context_data['page_can_create'], False)
        self.assertEqual(resp.context_data['page_can_edit'], True)

        with override_locale('de'):
            resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
            self.assertEqual(resp.context_data['page_slug'], sp.slug)
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(resp.context_data['page'].title, 'Title 2')
            self.assertIn('Body 2', resp.context_data['page'].body_html)
            self.assertEqual(resp.context_data['page_can_create'], False)
            self.assertEqual(resp.context_data['page_can_edit'], True)

        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': 'doesnotexist'}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page_slug'], 'doesnotexist')
        self.assertEqual(resp.context_data['page_can_create'], True)
        self.assertEqual(resp.context_data['page_can_edit'], False)

        # set up a wiki namespace
        ns = self.conf.wiki_namespaces.create(prefix='TEST', groups=['testgroup'])
        try:
            # fetch the namespaced page, expect it to be not existing and unable to create or edit
            resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': 'test-doesnotexist'}))
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(resp.context_data['page_slug'], 'test-doesnotexist')
            self.assertEqual(resp.context_data['page_can_create'], False)
            self.assertEqual(resp.context_data['page_can_edit'], False)

            # add the correct namespace to the user and try again, page should be creatable now
            try:
                self.conference_member.static_page_groups = ['testgroup']
                self.conference_member.save()
                resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': 'test-doesnotexist'}))
                self.assertEqual(resp.context_data['conf'], self.conf)
                self.assertEqual(resp.context_data['page_slug'], 'test-doesnotexist')
                self.assertEqual(resp.context_data['page_can_create'], True)
                self.assertEqual(resp.context_data['page_can_edit'], False)
            finally:
                self.conference_member.static_page_groups = []
                self.conference_member.save()
        finally:
            ns.delete()

        self.conf.require_ticket = False
        self.conf.save()
        ConferenceRequiredMixin._conf = None
        self.client.force_login(self.user_no_ticket)
        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp_private_conf.slug}))
        self.assertRedirects(resp, reverse('plainui:redeem_token'))

        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp_private_perm.slug}))
        self.assertEqual(resp.context_data['page'], None)
        self.assertEqual(resp.context_data['page_no_permission'], True)

        self.client.force_login(self.user)
        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp_private_conf.slug}))
        self.assertEqual(resp.context_data['page'], sp_private_conf)

        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp_private_perm.slug}))
        self.assertEqual(resp.context_data['page'], None)
        self.assertEqual(resp.context_data['page_no_permission'], True)

        self.conf.require_login = False
        self.conf.save()
        ConferenceRequiredMixin._conf = None
        self.client.logout()
        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['page'], sp)

        resp = self.client.get(reverse('plainui:static_page', kwargs={'page_slug': 'doesnotexist'}))
        self.assertEqual(resp.context_data['page_can_create'], False)

    @override_locale('en')
    def test_StaticPageEditView_get(self):
        from plainui.forms import StaticPageBodyForm

        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1')
        r1.save()
        r1.set_public()
        r2 = StaticPageRevision(title='Title 2', page=sp, body='Body 2')
        r2.save()
        r2.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title de1', page=sp_de, body='Body de1')
        r1_de.save()
        r1_de.set_public()
        r2_de = StaticPageRevision(title='Title de2', page=sp_de, body='Body de2')
        r2_de.save()
        r2_de.set_public()

        user2 = PlatformUser(username='testuser2', email='no2@where.test')
        user2.save()
        conference_member2 = ConferenceMember(conference=self.conf, user=user2, has_ticket=True)
        conference_member2.save()
        client2 = Client()
        client2.force_login(user2)

        self.assertNeedsLogin(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), check_user=True, check_conference_member=False)
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        lock = Lock.objects.get(content_type=ContentType.objects.get_for_model(StaticPage), object_id=sp.pk, lock_holder=self.user)
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['writeable'], True)
        self.assertEqual(resp.context_data['not_latest_revision'], False)
        self.assertEqual(resp.context_data['revision'], None)
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), r2.title)
        self.assertEqual(resp.context_data['form']['body'].value(), r2.body)
        self.assertEqual(resp.context_data['form']['lock'].value(), str(lock.pk))

        # user2 can't edit page as user1 has the lock
        resp = client2.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['lock_id'], '')
        self.assertNotEqual(resp.context_data['form']['lock'].value(), str(lock.pk))

        # selects page by language
        with override_locale('de'):
            resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'], sp_de)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['writeable'], True)
        self.assertEqual(resp.context_data['not_latest_revision'], False)
        self.assertEqual(resp.context_data['revision'], None)
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), r2_de.title)
        self.assertEqual(resp.context_data['form']['body'].value(), r2_de.body)

        # editing old revision
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'rev': r1.revision})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['writeable'], True)
        self.assertEqual(resp.context_data['not_latest_revision'], True)
        self.assertEqual(resp.context_data['revision'], str(r1.revision))
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), r1.title)
        self.assertEqual(resp.context_data['form']['body'].value(), r1.body)

        # readonly via page protection
        sp.protection = StaticPage.Protection.PERM
        sp.save()
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['writeable'], False)
        self.assertEqual(resp.context_data['lock_id'], '')
        self.assertEqual(resp.context_data['revision'], None)
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), r2.title)
        self.assertEqual(resp.context_data['form']['body'].value(), r2.body)
        self.assertEqual(resp.context_data['form'].fields['title'].disabled, True)
        self.assertEqual(resp.context_data['form'].fields['body'].disabled, True)

        # non-ticket user tries to edit a conference page
        sp.privacy = StaticPage.Privacy.CONFERENCE
        sp.save()
        self.client.force_login(self.user_no_ticket)
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        self.assertRedirects(resp, reverse('plainui:redeem_token'))
        self.assertSetsMessage(resp, 'You do not have the required permissions to edit this page.', level=messages.ERROR)

        # non-staff user tries to access an admin-accessible-only page
        sp.privacy = StaticPage.Privacy.PERM
        sp.save()
        self.client.force_login(self.user)
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
        self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
        self.assertSetsMessage(resp, 'You do not have the required permissions to edit this page.', level=messages.ERROR)

        # readonly via configured namespace
        sp.privacy = StaticPage.Privacy.NONE
        sp.protection = StaticPage.Protection.NONE
        sp.save()
        ns = self.conf.wiki_namespaces.create(prefix='TEST', groups=['testgroup'])
        try:
            resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(resp.context_data['page'], sp)
            self.assertEqual(resp.context_data['page_slug'], sp.slug)
            self.assertEqual(resp.context_data['writeable'], False)
            self.assertEqual(resp.context_data['lock_id'], '')
            self.assertEqual(resp.context_data['revision'], None)
            self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
            self.assertEqual(resp.context_data['form']['title'].value(), r2.title)
            self.assertEqual(resp.context_data['form']['body'].value(), r2.body)
            self.assertEqual(resp.context_data['form'].fields['title'].disabled, True)
            self.assertEqual(resp.context_data['form'].fields['body'].disabled, True)

            # namespace required but user has that namespace
            try:
                self.conference_member.static_page_groups = ['testgroup']
                self.conference_member.save()
                resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
                self.assertEqual(resp.context_data['page'], sp)
                self.assertEqual(resp.context_data['page_slug'], sp.slug)
                self.assertEqual(resp.context_data['writeable'], True)
                self.assertEqual(resp.context_data['revision'], None)
                self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
                self.assertEqual(resp.context_data['form']['title'].value(), r2.title)
                self.assertEqual(resp.context_data['form']['body'].value(), r2.body)
            finally:
                self.conference_member.static_page_groups = []
                self.conference_member.save()

            # namespace required, user doesn't have the required namespace but has static pages permission
            static_pages_group = Group.objects.get(name='Wiki Team')
            try:
                self.conference_member.is_staff = True
                self.conference_member.save()
                self.conference_member.permission_groups.add(static_pages_group)
                resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}))
                self.assertEqual(resp.context_data['page'], sp)
                self.assertEqual(resp.context_data['page_slug'], sp.slug)
                self.assertEqual(resp.context_data['writeable'], True)
                self.assertEqual(resp.context_data['revision'], None)
                self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
                self.assertEqual(resp.context_data['form']['title'].value(), r2.title)
                self.assertEqual(resp.context_data['form']['body'].value(), r2.body)
            finally:
                self.conference_member.permission_groups.remove(static_pages_group)
                self.conference_member.is_staff = False
                self.conference_member.save()

            # nonexistent page in namespace, not allowed to create but at least it doesn't crash
            with override_locale('en'):
                resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': 'test-doesnotexist'}))
                self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'test-doesnotexist'}))
                self.assertSetsMessage(resp, 'You do not have the required permissions to create this page.', level=messages.ERROR)
        finally:
            ns.delete()

        # nonexistent page, NOT in namespace -> show form with default values
        resp = self.client.get(reverse('plainui:static_page_edit', kwargs={'page_slug': 'test-doesnotexist'}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page_slug'], 'test-doesnotexist')
        self.assertEqual(resp.context_data['writeable'], True)
        self.assertEqual(resp.context_data['revision'], None)
        self.assertEqual(resp.context_data['lock_id'], '')  # no lock
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), 'test-doesnotexist')
        self.assertEqual(resp.context_data['form']['body'].value(), '')

    @override_locale('en')
    @override_settings(RATELIMIT_ENABLE=False)
    def test_StaticPageEditView_post(self):
        from plainui.forms import StaticPageBodyForm

        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1')
        r1.save()
        r1.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title de1', page=sp_de, body='Body de1')
        r1_de.save()
        r1_de.set_public()

        user2 = PlatformUser(username='testuser2', email='no2@where.test')
        user2.save()
        conference_member2 = ConferenceMember(conference=self.conf, user=user2, has_ticket=True)
        conference_member2.save()
        client2 = Client()
        client2.force_login(user2)

        self.assertNeedsLogin(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), post=True, check_user=True, check_conference_member=False)

        lock = Lock(content_type=ContentType.objects.get_for_model(StaticPage), object_id=sp.pk, lock_holder=self.user, timeout=timezone.now() + timedelta(30))
        lock.save()
        lock_id = str(lock.pk)

        # invalid input body
        resp = self.client.post(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title', 'body': '', 'lock': lock_id})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['writeable'], True)
        self.assertEqual(resp.context_data['revision'], None)
        self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
        self.assertEqual(resp.context_data['form']['title'].value(), 'New Title')
        self.assertEqual(resp.context_data['form']['body'].value(), '')
        self.assertTrue(resp.context_data['form'].errors)

        # invalid input title
        resp = self.client.post(reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': '', 'body': 'New Body', 'lock': lock_id})
        self.assertEqual(resp.context_data['form']['title'].value(), '')
        self.assertEqual(resp.context_data['form']['body'].value(), 'New Body')
        self.assertTrue(resp.context_data['form'].errors)

        # Mock the translations to stabelize the test for different locales
        with patch('plainui.views.static_pages.gettext', side_effect=mocktrans):
            # Valid input -> update page content and lock gets released
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title', 'body': 'New Body', 'lock': lock_id}
            )
            self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
            self.assertSetsMessage(resp, 'Updated Static Page_translated', level=messages.SUCCESS)
            sp.refresh_from_db()
            self.assertEqual(sp.revisions.count(), 2)
            self.assertEqual(sp.title, 'New Title')
            self.assertEqual(sp.body, 'New Body')
            self.assertEqual(sp_de.revisions.count(), 1)
            self.assertFalse(Lock.objects.filter(pk=lock.pk).exists())

            # Preview
            lock = Lock(
                content_type=ContentType.objects.get_for_model(StaticPage),
                object_id=sp.pk,
                lock_holder=self.user,
                timeout=timezone.now() + timedelta(seconds=30),
            )
            lock.save()
            lock_id = str(lock.pk)
            lock_timeout = lock.timeout
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}),
                {'preview': 'true', 'title': 'New Title Preview', 'body': 'New Body Preview', 'lock': lock_id},
            )
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(resp.context_data['page'], sp)
            self.assertEqual(resp.context_data['page_slug'], sp.slug)
            self.assertEqual(resp.context_data['writeable'], True)
            self.assertEqual(resp.context_data['revision'], None)
            self.assertIsInstance(resp.context_data['form'], StaticPageBodyForm)
            self.assertEqual(resp.context_data['form']['title'].value(), 'New Title Preview')
            self.assertEqual(resp.context_data['form']['body'].value(), 'New Body Preview')
            self.assertFalse(resp.context_data['form'].errors)
            self.assertEqual(resp.context_data['preview_title'], 'New Title Preview')
            self.assertIn('New Body Preview', resp.context_data['preview_body'])
            sp.refresh_from_db()
            self.assertEqual(sp.revisions.count(), 2)
            self.assertEqual(sp.title, 'New Title')
            self.assertEqual(sp.body, 'New Body')
            self.assertEqual(sp_de.revisions.count(), 1)
            lock.refresh_from_db()
            self.assertGreater(lock.timeout, lock_timeout)
            lock.delete()

            # page is protected -> normal user can't write it
            sp.protection = StaticPage.Protection.PERM
            sp.save()
            lock = Lock(
                content_type=ContentType.objects.get_for_model(StaticPage), object_id=sp.pk, lock_holder=self.user, timeout=timezone.now() + timedelta(30)
            )
            lock.save()
            lock_id = str(lock.pk)
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title2', 'body': 'New Body2', 'lock': lock_id}
            )
            self.assertEqual(resp.status_code, 403)
            sp.refresh_from_db()
            self.assertEqual(sp.revisions.count(), 2)
            self.assertEqual(sp.title, 'New Title')
            self.assertEqual(sp.body, 'New Body')
            self.assertEqual(sp_de.revisions.count(), 1)

            # page is in a namespace -> normal user can't write it
            sp.protection = StaticPage.Protection.NONE
            sp.save()
            ns = self.conf.wiki_namespaces.create(prefix='TEST', groups=['testgroup'])
            try:
                resp = self.client.post(
                    reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title3', 'body': 'New Body3', 'lock': lock_id}
                )
                self.assertEqual(resp.status_code, 403)
                sp.refresh_from_db()
                self.assertEqual(sp.revisions.count(), 2)
                self.assertEqual(sp.title, 'New Title')
                self.assertEqual(sp.body, 'New Body')
                self.assertEqual(sp_de.revisions.count(), 1)

                # but if user gets the namespace they can write
                # namespace required but user has that namespace
                try:
                    self.conference_member.static_page_groups = ['testgroup']
                    self.conference_member.save()

                    resp = self.client.post(
                        reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title4', 'body': 'New Body4', 'lock': lock_id}
                    )
                    self.assertSetsMessage(resp, 'Updated Static Page_translated', level=messages.SUCCESS)
                    self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
                    sp.refresh_from_db()
                    self.assertEqual(sp.revisions.count(), 3)
                    self.assertEqual(sp.title, 'New Title4')
                    self.assertEqual(sp.body, 'New Body4')
                    self.assertEqual(sp_de.revisions.count(), 1)
                finally:
                    self.conference_member.static_page_groups = []
                    self.conference_member.save()

                # namespace required, user doesn't have the required namespace but has static pages permission
                static_pages_group = Group.objects.get(name='Wiki Team')
                try:
                    self.conference_member.permission_groups.add(static_pages_group)
                    self.conference_member.is_staff = True
                    self.conference_member.save()
                    lock = Lock(
                        content_type=ContentType.objects.get_for_model(StaticPage),
                        object_id=sp.pk,
                        lock_holder=self.user,
                        timeout=timezone.now() + timedelta(30),
                    )
                    lock.save()
                    lock_id = str(lock.pk)
                    with patch('plainui.views.static_pages.gettext', side_effect=mocktrans):
                        resp = self.client.post(
                            reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'New Title5', 'body': 'New Body5', 'lock': lock_id}
                        )
                    self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': sp.slug}))
                    self.assertSetsMessage(resp, 'Updated Static Page_translated', level=messages.SUCCESS)
                    sp.refresh_from_db()
                    self.assertEqual(sp.revisions.count(), 4)
                    self.assertEqual(sp.title, 'New Title5')
                    self.assertEqual(sp.body, 'New Body5')
                    self.assertEqual(sp_de.revisions.count(), 1)
                finally:
                    self.conference_member.permission_groups.remove(static_pages_group)
                    self.conference_member.is_staff = False
                    self.conference_member.save()

                # can't create namespaced pages without permission
                resp = self.client.post(
                    reverse('plainui:static_page_edit', kwargs={'page_slug': 'test_new'}), {'title': 'Some New Page', 'body': 'Some New Page', 'lock': lock_id}
                )
                self.assertEqual(resp.status_code, 403)
                self.assertEqual(StaticPage.objects.filter(slug='test_new').count(), 0)
            finally:
                ns.delete()

            # page does not exist yet -> gets created
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': 'test_new'}), {'title': 'Some New Page!', 'body': 'Some New Page!', 'lock': ''}
            )
            self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'test_new'}))
            self.assertSetsMessage(resp, 'Created Static Page_translated', level=messages.SUCCESS)
            sp_new = StaticPage.objects.get(slug='test_new')
            self.assertEqual(sp_new.conference, self.conf)
            self.assertEqual(sp_new.language, None)
            self.assertEqual(sp_new.title, 'Some New Page!')
            self.assertEqual(sp_new.body, 'Some New Page!')
            spr = StaticPageRevision.objects.get(page=sp_new)
            self.assertEqual(spr.title, 'Some New Page!')
            self.assertEqual(spr.body, 'Some New Page!')
            self.assertEqual(spr.author, self.user)

            with override_settings(STATIC_PAGE_LOCALIZED_BY_DEFAULT=True):
                resp = self.client.post(
                    reverse('plainui:static_page_edit', kwargs={'page_slug': 'test_new2'}),
                    {'title': 'Some New Page!', 'body': 'Some New Page!', 'lock': lock_id},
                )
                self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'test_new2'}))

                sp_new2 = StaticPage.objects.get(slug='test_new2')
                self.assertEqual(sp_new2.conference, self.conf)
                self.assertEqual(sp_new2.language, 'en')

            # page links to nonexistent page, that then gets created, after which the link should be updated to refrect that the link target now exists
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': 'test_linking'}),
                {'title': 'Linking to other Pages test!', 'body': '[[test_linking_second_page]]'},
            )
            self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'test_linking'}))
            self.assertSetsMessage(resp, 'Created Static Page_translated', level=messages.SUCCESS)
            sp_new = StaticPage.objects.get(slug='test_linking')
            self.assertIn('internal-nonexist', sp_new.body_html)
            self.assertEqual(MarkdownMeta.objects.filter(src_type=ContentType.objects.get_for_model(StaticPage), src_object_id=sp_new.pk).count(), 1)
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': 'test_linking_second_page'}),
                {'title': 'Linked to test!', 'body': 'asdf', 'lock': lock_id},
            )
            sp_new.refresh_from_db()
            self.assertNotIn('internal-nonexist', sp_new.body_html)
            self.assertIn('internal', sp_new.body_html)
            self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'test_linking_second_page'}))

            # other user can't submit locked page
            Lock.objects.all().delete()
            lock = Lock(content_type=ContentType.objects.get_for_model(StaticPage), object_id=sp.pk, lock_holder=user2, timeout=timezone.now() + timedelta(30))
            lock.save()
            lock_timeout = lock.timeout
            lock_id = str(lock.pk)
            resp = self.client.post(
                reverse('plainui:static_page_edit', kwargs={'page_slug': sp.slug}), {'title': 'Title', 'body': 'Some Body', 'lock': uuid.uuid4()}
            )
            self.assertSetsMessage(resp, 'Lock error, could not save page!_translated', level=messages.ERROR)

            lock.refresh_from_db()
            self.assertEqual(lock.timeout, lock_timeout)

    @override_settings(RATELIMIT_ENABLE=False, STATIC_PAGE_LOCK_MAX_DURATION=60)
    def test_StaticPageLockKeepalive(self):
        user2 = PlatformUser(username='user2')
        user2.save()
        sp = StaticPage(conference=self.conf, slug='page', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1')
        r1.save()
        r1.set_public()
        lock = Lock(
            content_type=ContentType.objects.get_for_model(StaticPage),
            object_id=sp.pk,
            lock_holder=self.user,
            timeout=timezone.now() + timedelta(seconds=30),
        )
        lock.save()
        lock_timeout = lock.timeout

        self.assertNeedsLogin(reverse('plainui:static_page_refresh_lock'))

        # refreshes lock when called for existing page
        resp = self.client.post(reverse('plainui:static_page_refresh_lock'), {'page_slug': sp.slug, 'lock_id': str(lock.pk)})
        self.assertEqual(resp.status_code, 200)
        lock.refresh_from_db()
        self.assertGreater(lock.timeout, lock_timeout)
        lock_timeout = lock.timeout

        # does not refresh lock after max duration has expired
        with freeze_time(lock.created_at + timedelta(seconds=61)):
            resp = self.client.post(reverse('plainui:static_page_refresh_lock'), {'page_slug': sp.slug, 'lock_id': str(lock.pk)})
        self.assertEqual(resp.status_code, 200)
        lock.refresh_from_db()
        self.assertEqual(lock.timeout, lock_timeout)

        # does not refresh lock when called for nonexistent page
        resp = self.client.post(reverse('plainui:static_page_refresh_lock'), {'page_slug': 'something_else', 'lock_id': str(lock.pk)})
        self.assertEqual(resp.status_code, 200)
        lock.refresh_from_db()
        self.assertEqual(lock.timeout, lock_timeout)

        # does not refresh lock when called with bogus lock
        resp = self.client.post(reverse('plainui:static_page_refresh_lock'), {'page_slug': sp.slug, 'lock_id': str(uuid.uuid4())})
        self.assertEqual(resp.status_code, 200)
        lock.refresh_from_db()
        self.assertEqual(lock.timeout, lock_timeout)

        # does not refresh lock when lock is held by different user
        lock.lock_holder = user2
        lock.save()
        resp = self.client.post(reverse('plainui:static_page_refresh_lock'), {'page_slug': sp.slug, 'lock_id': str(lock.pk)})
        self.assertEqual(resp.status_code, 200)
        lock.refresh_from_db()
        self.assertEqual(lock.timeout, lock_timeout)

    @override_locale('en')
    def test_StaticPageHistoryView(self):
        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1', author=self.user)
        r1.save()
        r1.set_public()
        r2 = StaticPageRevision(title='Title 2', page=sp, body='Body 2', author=self.user)
        r2.save()
        r2.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title de1', page=sp_de, body='Body de1', author=self.user)
        r1_de.save()
        r1_de.set_public()

        self.assertNeedsLogin(reverse('plainui:static_page_history', kwargs={'page_slug': sp.slug}), check_user=True)

        resp = self.client.get(reverse('plainui:static_page_history', kwargs={'page_slug': sp.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertEqual(list(resp.context_data['history']), [r2, r1])

        with override_locale('de'):
            resp = self.client.get(reverse('plainui:static_page_history', kwargs={'page_slug': sp.slug}))
            self.assertEqual(resp.context_data['conf'], self.conf)
            self.assertEqual(resp.context_data['page_slug'], sp.slug)
            self.assertEqual(resp.context_data['page'], sp_de)
            self.assertEqual(list(resp.context_data['history']), [r1_de])

        resp = self.client.get(reverse('plainui:static_page_history', kwargs={'page_slug': 'doesnotexist'}))
        self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'doesnotexist'}))

    @override_locale('en')
    def test_StaticPageDiffView(self):
        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1', author=self.user)
        r1.save()
        r1.set_public()
        r2 = StaticPageRevision(title='Title 2', page=sp, body='Body 2', author=self.user)
        r2.save()
        r2.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title de1', page=sp_de, body='Body de1', author=self.user)
        r1_de.save()
        r1_de.set_public()

        self.assertNeedsLogin(reverse('plainui:static_page_diff', kwargs={'page_slug': sp.slug}), check_user=True)

        resp = self.client.get(reverse('plainui:static_page_diff', kwargs={'page_slug': sp.slug}), {'rev': r1.revision, 'rev2': r2.revision})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['page_slug'], sp.slug)
        self.assertEqual(resp.context_data['page'], sp)
        self.assertTrue(resp.context_data['diff'])

        resp = self.client.get(reverse('plainui:static_page_diff', kwargs={'page_slug': 'doesnotexist'}), {'rev': '1', 'rev2': '2'})
        self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'doesnotexist'}))

    @freeze_time(datetime(2020, 1, 1, 1, 0, 0, tzinfo=UTC))
    @override_locale('en')
    def test_StaticPageGlobalHistoryView(self):
        sp = StaticPage(conference=self.conf, slug='test1', language='en')
        sp.save()
        r1 = StaticPageRevision(title='Title 1', page=sp, body='Body 1', author=self.user)
        r1.save()
        r1.set_public()
        r2 = StaticPageRevision(title='Title 2', page=sp, body='Body 2', author=self.user)
        r2.save()
        r2.set_public()

        sp2 = StaticPage(conference=self.conf, slug='test2', language='en')
        sp2.save()
        r21 = StaticPageRevision(title='Title 21', page=sp, body='Body 21', author=self.user)
        r21.save()
        r21.set_public()

        sp_de = StaticPage(conference=self.conf, slug='test1', language='de')
        sp_de.save()
        r1_de = StaticPageRevision(title='Title de1', page=sp_de, body='Body de1', author=self.user)
        r1_de.save()
        r1_de.set_public()

        self.assertNeedsLogin(reverse('plainui:static_page_global_history'), check_user=True)

        resp = self.client.get(reverse('plainui:static_page_global_history'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['history'], {'2020-01-01': [r21, r2, r1]})

        resp = self.client.get(reverse('plainui:static_page_global_history'), {'page': '0'})
        self.assertEqual(resp.context_data['history'], {'2020-01-01': [r21, r2, r1]})

        resp = self.client.get(reverse('plainui:static_page_global_history'), {'page': '1'})
        self.assertEqual(resp.context_data['history'], {})

        with override_locale('de'):
            resp = self.client.get(reverse('plainui:static_page_global_history'))
            self.assertEqual(resp.context_data['history'], {'2020-01-01': [r1_de]})

    def test_StaticPageRedirectToStart(self):
        self.assertNeedsLogin(reverse('plainui:static_page_redirect_to_start'))
        resp = self.client.get(reverse('plainui:static_page_redirect_to_start'), follow=False)
        self.assertRedirects(resp, reverse('plainui:static_page', kwargs={'page_slug': 'start'}))

    @override_locale('en')
    def test_SearchView_get(self):
        self.assertNeedsLogin(reverse('plainui:search'))
        resp = self.client.get(reverse('plainui:search'), follow=False)
        self.assertRedirects(resp, reverse('plainui:index'))

    @override_locale('en')
    def test_SearchView_post(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='asdf', state=Assembly.State.PLACED)
        assembly.save()
        event = Event(
            conference=self.conf,
            assembly=assembly,
            name='asdf',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event.save()
        tag = ConferenceTag(conference=self.conf, slug='aSdf', is_public=True)
        tag.save()
        track = ConferenceTrack(conference=self.conf, slug='asdf', name='asdf')
        track.save()
        page = StaticPage(conference=self.conf, slug='page', language='en')
        page.save()
        revision = StaticPageRevision(page=page, title='asdf', body='foo')
        revision.save()
        revision.set_public()

        self.assertNeedsLogin(reverse('plainui:search'), data={'q': 'asdf'}, post=True)
        resp = self.client.post(reverse('plainui:search'), {'q': 'asdf'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['search_query'], 'asdf')
        self.assertEqual(
            sorted(resp.context_data['search_results'], key=lambda r: r['type']),
            [
                {'type': 'Assembly', 'item': assembly},
                {'type': 'ConferenceTag', 'item': tag},
                {'type': 'ConferenceTrack', 'item': track},
                {'type': 'Event', 'item': event},
                {'type': 'StaticPage', 'item': page},
            ],
        )

    @override_locale('en')
    def test_ProfileView_get(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        event = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_2',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 1, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()
        badge = Badge(conference=self.conf, name='badge-name', issuing_assembly=assembly)
        badge.save()

        self.assertNeedsLogin(reverse('plainui:userprofile'), check_user=True)
        resp = self.client.get(reverse('plainui:userprofile'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['badges']), [])
        self.assertEqual(list(resp.context_data['my_favorite_events']), [])
        self.assertEqual(list(resp.context_data['my_favorite_assemblies']), [])
        self.assertEqual(list(resp.context_data['is_favorite_events']), [])
        self.assertEqual(list(resp.context_data['is_favorite_assemblies']), [])

        self.assertIsInstance(resp.context_data['form'], ProfileEditForm)

        self.user.favorite_events.add(event)
        self.user.favorite_assemblies.add(assembly)
        session = self.client.session
        session['fav_e'] = session['fav_a'] = session['sch_e'] = None
        session.save()

        user_badge = UserBadge(user=self.user, badge=badge, accepted_by_user=True, visibility=UserBadgeVisibility.PUBLIC)
        user_badge.acceptByUser()
        user_badge.save()
        resp = self.client.get(reverse('plainui:userprofile'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['badges']), [user_badge])
        self.assertEqual(list(resp.context_data['my_favorite_events']), [event])
        self.assertEqual(list(resp.context_data['my_favorite_assemblies']), [assembly])
        self.assertEqual(list(resp.context_data['is_favorite_events']), [str(event.pk)])
        self.assertEqual(list(resp.context_data['is_favorite_assemblies']), [str(assembly.pk)])

    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    @patch('modeltranslation.settings.AVAILABLE_LANGUAGES', ['en', 'de'])
    def test_ProfileView_post(self):
        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        event = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event.save()

        self.assertNeedsLogin(reverse('plainui:userprofile'), post=True, check_user=True)
        resp = self.client.post(
            reverse('plainui:userprofile'),
            {
                'description_de': 'new_description de',
                'description_en': 'new_description en',
                'default_badge_visibility': 'private',
                'pronouns': 'they',
                'timezone': 'Europe/Berlin',
            },
        )
        self.assertRedirects(resp, reverse('plainui:userprofile'))
        self.user.refresh_from_db()
        self.conference_member.refresh_from_db()
        self.assertEqual(self.conference_member.description_de, 'new_description de')
        self.assertEqual(self.conference_member.description_en, 'new_description en')
        self.assertEqual(self.user.pronouns, 'they')
        self.assertEqual(self.user.timezone.key, 'Europe/Berlin')
        self.assertSetsMessage(resp, 'Updated Profile')

    @override_locale('en')
    def test_ModifyThemeView_get(self):
        self.assertNeedsLogin(reverse('plainui:modify_theme'))
        resp = self.client.get(reverse('plainui:modify_theme'), follow=False)
        self.assertRedirects(resp, reverse('plainui:index'))

    @override_locale('en')
    def test_ModifyThemeView_post(self):
        self.assertNeedsLogin(reverse('plainui:modify_theme'), post=True)

        # setting high_contrast works
        resp = self.client.post(
            reverse('plainui:modify_theme'),
            {
                'high_contrast': 'True',
                'next': reverse('plainui:landing'),
            },
        )
        self.assertRedirects(resp, reverse('plainui:landing'))
        self.user.refresh_from_db()
        self.assertEqual(self.user.high_contrast, True)
        self.assertEqual(self.client.session['high_contrast'], True)

        # invalid destination redirects to index
        resp = self.client.post(
            reverse('plainui:modify_theme'),
            {
                'high_contrast': 'True',
                'next': 'https://www.google.de/',
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.user.refresh_from_db()
        self.assertEqual(self.user.high_contrast, True)
        self.assertEqual(self.client.session['high_contrast'], True)

        # invalid high_contrast defaults to False
        resp = self.client.post(
            reverse('plainui:modify_theme'),
            {
                'high_contrast': 'doesnotexist',
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.user.refresh_from_db()
        self.assertEqual(self.user.high_contrast, False)
        self.assertEqual(self.client.session['high_contrast'], False)

    @override_settings(LANGUAGE_CODE='en', PRETIX_SECRET_KEY='^v^', PRETIX_ISSUER='tickets.events.ccc.de')
    @override_locale('en')
    @freeze_time(datetime(2020, 12, 28, 1, 0, 0, tzinfo=UTC))
    def test_RedeemTokenView(self):
        self.conf.start = datetime(2020, 12, 27, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()
        valid_token = jwt.encode(
            {
                'aud': self.conf.slug,
                'exp': datetime.now(UTC) + timedelta(hours=1),
                'iat': datetime.now(UTC) - timedelta(hours=1),
                'iss': 'tickets.events.ccc.de',
                'uid': 'blblbl123',
            },
            key='^v^',
        )
        invalid_token = (
            'fyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhdWQiOiJzbHVnMSIsImV4cCI6MTYwODc2ODMwNywiaWF0IjoxNjA'
            '4NzYxMTA3LCJpc3MiOiJ0aWNrZXRzLmV2ZW50cy5jY2MuZGUiLCJ1aWQiOiJibGJsYmwxMjMifQ.QglU9dLHbIAOJ2e8hj_HSc6rZBpAcTrPbvY2H7HXnFw'
        )

        resp = self.client.get(reverse('plainui:redeem_token'))
        self.assertRedirects(resp, reverse('plainui:index'))

        ConferenceMember.objects.all().delete()
        self.client.logout()  # reset conferencemember cache in session
        self.client.force_login(self.user)
        resp = self.client.get(reverse('plainui:redeem_token'), {'jwt': valid_token})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'authenticated')

        self.client.logout()
        resp = self.client.get(reverse('plainui:redeem_token'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'token')
        self.assertTrue(resp.context_data['form'])

        # bullshit input
        resp = self.client.get(reverse('plainui:redeem_token'), {'jwt': 'bullshit'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'token')
        self.assertTrue(resp.context_data['form'])
        self.assertTrue(resp.context_data['form'].errors)

        # invalid token input
        resp = self.client.get(reverse('plainui:redeem_token'), {'jwt': invalid_token})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'token')
        self.assertTrue(resp.context_data['form'])
        self.assertTrue(resp.context_data['form'].errors)

        # valid token input
        resp = self.client.get(reverse('plainui:redeem_token'), {'jwt': valid_token})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertTrue(resp.context_data['form_add'])
        self.assertTrue(resp.context_data['form_create'])

        # conference over
        self.conf.end = datetime(2020, 12, 28, 0, 0, 0, tzinfo=UTC)
        self.conf.save()
        resp = self.client.get(reverse('plainui:redeem_token'), {'jwt': valid_token})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'token')
        self.assertTrue(resp.context_data['form'])
        self.assertTrue(resp.context_data['form'].errors)

    @override_settings(LANGUAGE_CODE='en', PRETIX_SECRET_KEY='^v^', PRETIX_ISSUER='tickets.events.ccc.de', RATELIMIT_ENABLE=False)
    @override_locale('en')
    @freeze_time(datetime(2020, 12, 28, 1, 0, 0, tzinfo=UTC))
    def test_RedeemTokenAddToUserView(self):
        self.conf.start = datetime(2020, 12, 27, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()
        self.client.logout()
        self.user.set_password('test')
        self.user.save()
        ConferenceMember.objects.all().delete()
        valid_token = jwt.encode(
            {
                'aud': self.conf.slug,
                'exp': datetime.now(UTC) + timedelta(hours=1),
                'iat': datetime.now(UTC) - timedelta(hours=1),
                'iss': 'tickets.events.ccc.de',
                'uid': 'blblbl123',
            },
            key='^v^',
        )

        # get-redirect
        resp = self.client.get(reverse('plainui:redeem_token_add_to_user'))
        self.assertRedirects(resp, reverse('plainui:redeem_token'))

        # invalid token
        invalid_token = (
            'fyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhdWQiOiJzbHVnMSIsImV4cCI6MTYwODc2ODMwNywiaWF0IjoxNjA'
            '4NzYxMTA3LCJpc3MiOiJ0aWNrZXRzLmV2ZW50cy5jY2MuZGUiLCJ1aWQiOiJibGJsYmwxMjMifQ.QglU9dLHbIAOJ2e8hj_HSc6rZBpAcTrPbvY2H7HXnFw'
        )
        with patch('core.models.ticket._', side_effect=mocktrans):
            resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': invalid_token, 'username': self.user.username, 'password': 'test'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertEqual(list(resp.context_data['form_add'].non_field_errors()), ['ConferenceMemberTicket__token_error_translated'])
        self.assertTrue(resp.context_data['form_create'])

        # conference already over
        self.conf.end = datetime(2020, 12, 28, 0, 0, 0, tzinfo=UTC)
        self.conf.save()
        resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': valid_token, 'username': self.user.username, 'password': 'test'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertEqual(list(resp.context_data['form_add'].non_field_errors()), ['The Conference is over!'])
        self.assertTrue(resp.context_data['form_create'])
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()

        # submit valid token and join conference
        resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': valid_token, 'username': self.user.username, 'password': 'test'})
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertTrue(ConferenceMemberTicket.objects.filter(conference=self.conf, user=self.user, ident='blblbl123').exists())
        self.assertTrue(ConferenceMember.objects.filter(conference=self.conf, user=self.user).exists())

        # user already has a ConferenceMember but no ticket -> add ticket but don't create second ConferenceMember
        ConferenceMemberTicket.objects.all().delete()
        self.assertEqual(ConferenceMember.objects.filter(conference=self.conf, user=self.user).count(), 1)
        resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': valid_token, 'username': self.user.username, 'password': 'test'})
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertTrue(ConferenceMemberTicket.objects.filter(conference=self.conf, user=self.user, ident='blblbl123').exists())
        self.assertEqual(ConferenceMember.objects.filter(conference=self.conf, user=self.user).count(), 1)

        # testing the rate limiting: Will block after 5 allowed requests
        with override_settings(RATELIMIT_ENABLE=True):
            for i in range(5):
                ConferenceMemberTicket.objects.all().delete()
                ConferenceMember.objects.all().delete()
                resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': valid_token, 'username': self.user.username, 'password': 'test'})
                self.assertRedirects(resp, reverse('plainui:index'))

            ConferenceMemberTicket.objects.all().delete()
            ConferenceMember.objects.all().delete()
            resp = self.client.post(reverse('plainui:redeem_token_add_to_user'), {'token': valid_token, 'username': self.user.username, 'password': 'test'})
            self.assertIsInstance(resp, HttpResponseRateLimited)

    @override_settings(LANGUAGE_CODE='en', PRETIX_SECRET_KEY='^v^', PRETIX_ISSUER='tickets.events.ccc.de', AUTH_PASSWORD_VALIDATORS=[])
    @override_locale('en')
    @freeze_time(datetime(2020, 12, 28, 1, 0, 0, tzinfo=UTC))
    def test_RedeemTokenUserCreateView(self):
        self.conf.start = datetime(2020, 12, 27, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()
        self.client.logout()
        invalid_token = (
            'fyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhdWQiOiJzbHVnMSIsImV4cCI6MTYwODc2ODMwNywiaWF0IjoxNjA'
            '4NzYxMTA3LCJpc3MiOiJ0aWNrZXRzLmV2ZW50cy5jY2MuZGUiLCJ1aWQiOiJibGJsYmwxMjMifQ.QglU9dLHbIAOJ2e8hj_HSc6rZBpAcTrPbvY2H7HXnFw'
        )
        valid_token = jwt.encode(
            {
                'aud': self.conf.slug,
                'exp': datetime.now(UTC) + timedelta(hours=1),
                'iat': datetime.now(UTC) - timedelta(hours=1),
                'iss': 'tickets.events.ccc.de',
                'uid': 'blblbl123',
            },
            key='^v^',
        )

        # get-redirect
        resp = self.client.get(reverse('plainui:redeem_token_create_user'))
        self.assertRedirects(resp, reverse('plainui:redeem_token'))

        # invalid token
        with patch('core.models.ticket._', side_effect=mocktrans):
            resp = self.client.post(
                reverse('plainui:redeem_token_create_user'), {'token': invalid_token, 'username': 'testuser2', 'password1': 'test', 'password2': 'test'}
            )
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertTrue(resp.context_data['form_add'])
        self.assertEqual(list(resp.context_data['form_create'].non_field_errors()), ['ConferenceMemberTicket__token_error_translated'])

        # submit valid token but username exists already
        resp = self.client.post(
            reverse('plainui:redeem_token_create_user'), {'token': valid_token, 'username': self.user.username, 'password1': 'test', 'password2': 'test'}
        )
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertTrue(resp.context_data['form_add'])
        self.assertEqual(list(resp.context_data['form_create']['username'].errors), ['A user with that username already exists.'])

        # valid token but conference already over
        self.conf.end = datetime(2020, 12, 28, 0, 0, 0, tzinfo=UTC)
        self.conf.save()
        resp = self.client.post(
            reverse('plainui:redeem_token_create_user'), {'token': valid_token, 'username': self.user.username, 'password1': 'test', 'password2': 'test'}
        )
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['step'], 'user')
        self.assertTrue(resp.context_data['form_add'])
        self.assertEqual(list(resp.context_data['form_create'].non_field_errors()), ['The Conference is over!'])
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()

        # submit valid token and join conference
        resp = self.client.post(
            reverse('plainui:redeem_token_create_user'), {'token': valid_token, 'username': 'testuser2', 'password1': 'test', 'password2': 'test'}
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        user = PlatformUser.objects.get(username='testuser2')
        self.assertTrue(ConferenceMemberTicket.objects.filter(conference=self.conf, user=user, ident='blblbl123').exists())
        self.assertTrue(ConferenceMember.objects.filter(conference=self.conf, user=user).exists())

    @override_settings(LANGUAGE_CODE='en', PRETIX_SECRET_KEY='^v^', PRETIX_ISSUER='tickets.events.ccc.de', AUTH_PASSWORD_VALIDATORS=[])
    @override_locale('en')
    @freeze_time(datetime(2020, 12, 28, 1, 0, 0, tzinfo=UTC))
    def test_RedeemTokenLoggedIn(self):
        self.conf.start = datetime(2020, 12, 27, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()
        ConferenceMember.objects.all().delete()
        invalid_token = (
            'fyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhdWQiOiJzbHVnMSIsImV4cCI6MTYwODc2ODMwNywiaWF0IjoxNjA'
            '4NzYxMTA3LCJpc3MiOiJ0aWNrZXRzLmV2ZW50cy5jY2MuZGUiLCJ1aWQiOiJibGJsYmwxMjMifQ.QglU9dLHbIAOJ2e8hj_HSc6rZBpAcTrPbvY2H7HXnFw'
        )
        valid_token = jwt.encode(
            {
                'aud': self.conf.slug,
                'exp': datetime.now(UTC) + timedelta(hours=1),
                'iat': datetime.now(UTC) - timedelta(hours=1),
                'iss': 'tickets.events.ccc.de',
                'uid': 'blblbl123',
            },
            key='^v^',
        )

        self.assertNeedsLogin(reverse('plainui:redeem_token_loggedin'), check_conference_member=False, check_user=True)
        resp = self.client.get(reverse('plainui:redeem_token_loggedin'))
        self.assertRedirects(resp, reverse('plainui:redeem_token'))

        # invalid token
        self.conf.end = datetime(2020, 12, 28, 0, 0, 0, tzinfo=UTC)
        self.conf.save()
        resp = self.client.post(reverse('plainui:redeem_token_loggedin'), {'token': invalid_token})
        self.assertRedirects(resp, reverse('plainui:redeem_token'))
        self.assertSetsMessage(resp, 'The Conference is over!', level=messages.ERROR)
        self.conf.end = datetime(2020, 12, 31, 23, 59, 0, tzinfo=UTC)
        self.conf.save()

        # valid token but conference over
        with patch('core.models.ticket._', side_effect=mocktrans):
            resp = self.client.post(reverse('plainui:redeem_token_loggedin'), {'token': invalid_token})
        self.assertRedirects(resp, reverse('plainui:redeem_token'))
        self.assertSetsMessage(resp, 'ConferenceMemberTicket__token_error_translated', level=messages.ERROR)

        # valid token
        resp = self.client.post(reverse('plainui:redeem_token_loggedin'), {'token': valid_token})
        self.assertTrue(ConferenceMemberTicket.objects.filter(conference=self.conf, user=self.user, ident='blblbl123').exists())
        self.assertTrue(ConferenceMember.objects.filter(conference=self.conf, user=self.user).exists())
        self.assertRedirects(resp, reverse('plainui:index'))

    @override_settings(LANGUAGE_CODE='en', PRETIX_SECRET_KEY='^v^', PRETIX_ISSUER='tickets.events.ccc.de', AUTH_PASSWORD_VALIDATORS=[])
    @override_locale('en')
    def test_TokenPasswordResetView(self):
        ticket_data = {
            'aud': self.conf.slug,
            'exp': datetime.now(UTC) + timedelta(hours=1),
            'iat': datetime.now(UTC) - timedelta(hours=1),
            'iss': 'tickets.events.ccc.de',
        }
        valid_token = jwt.encode({'uid': 'blblbl123', **ticket_data}, key='^v^')
        decoded_token = jwt.decode(valid_token, key='^v^', audience=self.conf.slug, issuer='tickets.events.ccc.de', algorithms=['HS256'])
        del decoded_token['uid']
        ConferenceMemberTicket(user=self.user, ident='blblbl123', conference=self.conf, data=decoded_token).save()
        self.user.set_password('asdf')
        self.user.email = ''
        self.user.save()

        # invalid token
        invalid_token = (
            'fyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhdWQiOiJzbHVnMSIsImV4cCI6MTYwODc2ODMwNywiaWF0IjoxNjA'
            '4NzYxMTA3LCJpc3MiOiJ0aWNrZXRzLmV2ZW50cy5jY2MuZGUiLCJ1aWQiOiJibGJsYmwxMjMifQ.QglU9dLHbIAOJ2e8hj_HSc6rZBpAcTrPbvY2H7HXnFw'
        )
        resp = self.client.post(
            reverse('plainui:token_password_reset'),
            {
                'jwt': invalid_token,
                'username': self.user.username,
                'new_password1': 'new',
                'new_password2': 'new',
            },
        )
        self.assertTrue(resp.context_data['form'].errors)
        self.user.refresh_from_db()
        self.assertFalse(self.user.check_password('new'))

        resp = self.client.post(
            reverse('plainui:token_password_reset'),
            {
                'jwt': valid_token,
                'username': 'doesnotexist',
                'new_password1': 'new',
                'new_password2': 'new',
            },
        )
        self.assertTrue(resp.context_data['form'].errors)
        self.user.refresh_from_db()
        self.assertFalse(self.user.check_password('new'))

        resp = self.client.post(
            reverse('plainui:token_password_reset'),
            {
                'jwt': valid_token,
                'username': self.user.username,
                'new_password1': 'new',
                'new_password2': 'new2',
            },
        )
        self.assertTrue(resp.context_data['form'].errors)
        self.user.refresh_from_db()
        self.assertFalse(self.user.check_password('new'))

        resp = self.client.post(
            reverse('plainui:token_password_reset'),
            {
                'jwt': valid_token,
                'username': self.user.username,
                'new_password1': 'new',
                'new_password2': 'new',
            },
        )
        self.assertRedirects(resp, reverse('plainui:login'))
        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password('new'))
        self.assertSetsMessage(resp, 'Passwort successfully reset!')

    @override_locale('en')
    def test_ModifyFavoritesView_get(self):
        self.assertNeedsLogin(reverse('plainui:modify_favorites'), check_user=True)
        resp = self.client.get(reverse('plainui:modify_favorites'))
        self.assertRedirects(resp, reverse('plainui:userprofile'))

    @override_locale('en')
    def test_ModifyFavoritesView_post(self):
        assembly1 = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly1.save()
        assembly2 = Assembly(conference=self.conf, slug='assembly2', name='Assembly2', state=Assembly.State.PLACED)
        assembly2.save()
        event1 = Event(
            conference=self.conf,
            assembly=assembly1,
            name='Event1_1',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event1.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly1,
            name='Event1_2',
            is_public=True,
            schedule_start=datetime(2020, 1, 2, 1, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()

        self.assertNeedsLogin(reverse('plainui:modify_favorites'), post=True, check_user=True)
        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'add',
                'type': 'assembly',
                'id': str(assembly1.pk),
            },
        )
        self.assertRedirects(resp, reverse('plainui:userprofile'))
        self.assertTrue(self.user.favorite_assemblies.filter(pk=assembly1.pk).exists())
        self.assertEqual(self.client.session['fav_a'], [str(assembly1.pk)])

        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'add',
                'type': 'assembly',
                'id': str(assembly2.pk),
            },
        )
        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'remove',
                'type': 'assembly',
                'id': str(assembly1.pk),
                'next': reverse('plainui:index'),
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertFalse(self.user.favorite_assemblies.filter(pk=assembly1.pk).exists())
        self.assertEqual(self.client.session['fav_a'], [str(assembly2.pk)])

        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'add',
                'type': 'event',
                'id': str(event1.pk),
                'next': reverse('plainui:index'),
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertTrue(self.user.favorite_events.filter(pk=event1.pk).exists())
        self.assertEqual(self.client.session['fav_e'], [str(event1.pk)])

        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'add',
                'type': 'event',
                'id': str(event2.pk),
                'next': reverse('plainui:index'),
            },
        )

        resp = self.client.post(
            reverse('plainui:modify_favorites'),
            {
                'mode': 'remove',
                'type': 'event',
                'id': str(event1.pk),
            },
        )
        self.assertRedirects(resp, reverse('plainui:userprofile'))
        self.assertFalse(self.user.favorite_events.filter(pk=event1.pk).exists())
        self.assertEqual(self.client.session['fav_e'], [str(event2.pk)])

    @override_locale('en')
    def test_PersonalMessageListView(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()
        dm_recv1 = DirectMessage(conference=self.conf, sender=user2, recipient=self.user, subject='sj1', body='bd1')
        dm_recv1.save()
        dm_recv2 = DirectMessage(conference=self.conf, sender=user2, recipient=self.user, subject='sj2', body='bd2')
        dm_recv2.save()
        dm_sent = DirectMessage(conference=self.conf, sender=self.user, recipient=user2, subject='sj3', body='bd3')
        dm_sent.save()

        self.assertNeedsLogin(reverse('plainui:personal_message'), check_user=True)
        resp = self.client.get(reverse('plainui:personal_message'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        msgs = list(resp.context_data['msgs'])
        self.assertEqual(msgs, [dm_recv2, dm_recv1])
        self.assertEqual(msgs[0].sender, user2)
        self.assertEqual(resp.context_data['total'], 2)

        resp = self.client.get(reverse('plainui:personal_message'), {'start': 1})
        self.assertEqual(resp.context_data['conf'], self.conf)
        msgs = list(resp.context_data['msgs'])
        self.assertEqual(msgs, [dm_recv1])
        self.assertEqual(msgs[0].sender, user2)
        self.assertEqual(resp.context_data['total'], 2)

        resp = self.client.get(reverse('plainui:personal_message_outbox'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        msgs = list(resp.context_data['msgs'])
        self.assertEqual(msgs, [dm_sent])
        self.assertEqual(msgs[0].recipient, user2)
        self.assertEqual(resp.context_data['total'], 1)

    @override_locale('en')
    def test_PersonalMessageSendView_get(self):
        self.assertNeedsLogin(reverse('plainui:personal_message_send'), check_user=True)
        resp = self.client.get(reverse('plainui:personal_message_send'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['form']['recipient'].value(), '')

        resp = self.client.get(reverse('plainui:personal_message_send_to', kwargs={'recipient': 'blgl'}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['form']['recipient'].value(), 'blgl')

    @override_settings(LANGUAGE_CODE='en', RATELIMIT_ENABLE=False)
    @override_locale('en')
    def test_PersonalMessageSendView_post(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()

        self.assertNeedsLogin(reverse('plainui:personal_message_send'), post=True, check_user=True)
        resp = self.client.post(
            reverse('plainui:personal_message_send'),
            {
                'in_reply_to': '',
                'recipient': 'testuser2',
                'subject': 'Subj',
                'body': 'Body',
            },
        )
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        new_dm = DirectMessage.objects.get()
        self.assertEqual(new_dm.conference, self.conf)
        self.assertEqual(new_dm.sender, self.user)
        self.assertEqual(new_dm.recipient, user2)
        self.assertEqual(new_dm.subject, 'Subj')
        self.assertEqual(new_dm.body, 'Body')
        self.assertSetsMessage(resp, 'Message sent.')

        received_dm = DirectMessage(conference=self.conf, sender=user2, recipient=self.user)
        received_dm.save()
        resp = self.client.post(
            reverse('plainui:personal_message_send'),
            {
                'in_reply_to': str(received_dm.pk),
                'recipient': 'testuser2',
                'subject': 'Subj2',
                'body': 'Body2',
            },
        )
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        new_dm = DirectMessage.objects.get(in_reply_to=received_dm)
        self.assertEqual(new_dm.conference, self.conf)
        self.assertEqual(new_dm.sender, self.user)
        self.assertEqual(new_dm.recipient, user2)
        self.assertEqual(new_dm.subject, 'Subj2')
        self.assertEqual(new_dm.body, 'Body2')
        self.assertEqual(new_dm.in_reply_to, received_dm)
        received_dm.refresh_from_db()
        self.assertTrue(received_dm.has_responded)
        self.assertSetsMessage(resp, 'Message sent.')

        resp = self.client.post(
            reverse('plainui:personal_message_send'),
            {
                'in_reply_to': '',
                'recipient': 'Unknown User',
                'subject': '',
                'body': '',
            },
        )
        form = resp.context_data['form']
        self.assertEqual(form.errors['recipient'], ['Unknown User!'])
        self.assertEqual(form.errors['subject'], ['This field is required.'])
        self.assertEqual(form.errors['body'], ['This field is required.'])

        DirectMessage.objects.all().delete()
        self.user.shadow_banned = True
        self.user.save()
        resp = self.client.post(
            reverse('plainui:personal_message_send'),
            {
                'in_reply_to': '',
                'recipient': 'testuser2',
                'subject': 'subject',
                'body': 'body',
            },
        )
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        new_dm = DirectMessage.objects.get(subject='subject')
        self.user.refresh_from_db()
        self.assertTrue(new_dm.deleted_by_recipient, True)

        # testing the rate limiting: Will block after 3 allowed requests
        self.user.shadow_banned = False
        self.user.save()
        with override_settings(RATELIMIT_ENABLE=True):
            for i in range(3):
                resp = self.client.post(
                    reverse('plainui:personal_message_send'),
                    {
                        'in_reply_to': '',
                        'recipient': 'testuser2',
                        'subject': 'subject' + str(i),
                        'body': 'body',
                    },
                )
                self.assertRedirects(resp, reverse('plainui:personal_message'))

            resp = self.client.post(
                reverse('plainui:personal_message_send'),
                {
                    'in_reply_to': '',
                    'recipient': 'testuser2',
                    'subject': 'subject-ratelimited',
                    'body': 'body-ratelimited',
                },
            )
            self.assertIsInstance(resp, HttpResponseForbidden)

    @override_locale('en')
    def test_PersonalMessageShowView(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()
        dm_recv = DirectMessage(conference=self.conf, sender=user2, recipient=self.user, subject='SomeSubj', body='Some*Bdy*', was_read=False)
        dm_recv.save()
        dm_sent = DirectMessage(conference=self.conf, sender=self.user, recipient=user2, subject='OtherSubj', body='**OtherBdy**', was_read=False)
        dm_sent.save()

        self.assertNeedsLogin(reverse('plainui:personal_message_show', kwargs={'msg_id': str(dm_recv.pk)}), check_user=True)
        resp = self.client.get(reverse('plainui:personal_message_show', kwargs={'msg_id': str(dm_recv.pk)}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['msg'], dm_recv)
        self.assertEqual(str(resp.context_data['msg_body']), str(render_markdown(self.conf, dm_recv.body)))
        dm_recv.refresh_from_db()
        self.assertTrue(dm_recv.was_read)

        resp = self.client.get(reverse('plainui:personal_message_show', kwargs={'msg_id': str(dm_sent.pk)}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['msg'], dm_sent)
        self.assertEqual(str(resp.context_data['msg_body']), str(render_markdown(self.conf, dm_sent.body)))
        dm_sent.refresh_from_db()
        self.assertFalse(dm_sent.was_read)

    @override_locale('en')
    def test_PersonalMessageDeleteView_get(self):
        self.assertNeedsLogin(reverse('plainui:personal_message_delete'), check_user=True)
        resp = self.client.get(reverse('plainui:personal_message_delete'))
        self.assertRedirects(resp, reverse('plainui:personal_message'))

    @override_settings(LANGUAGE_CODE='en')
    @override_locale('en')
    def test_PersonalMessageDeleteView_post(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()
        dm = DirectMessage(conference=self.conf, sender=user2, recipient=self.user)
        dm.save()
        dm_halfdel1 = DirectMessage(conference=self.conf, sender=user2, recipient=self.user, deleted_by_sender=True)
        dm_halfdel1.save()
        dm_halfdel2 = DirectMessage(conference=self.conf, sender=self.user, recipient=user2, deleted_by_recipient=True)
        dm_halfdel2.save()

        self.assertNeedsLogin(reverse('plainui:personal_message_delete'), post=True, check_user=True)
        resp = self.client.post(reverse('plainui:personal_message_delete'), {'id': str(dm.pk)})
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        dm.refresh_from_db()
        self.assertTrue(dm.deleted_by_recipient)
        self.assertFalse(dm.deleted_by_sender)

        resp = self.client.post(reverse('plainui:personal_message_delete'), {'id': str(dm_halfdel1.pk)})
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        self.assertFalse(DirectMessage.objects.filter(pk=dm_halfdel1.pk).exists())

        resp = self.client.post(reverse('plainui:personal_message_delete'), {'id': str(dm_halfdel2.pk)})
        self.assertRedirects(resp, reverse('plainui:personal_message'))
        self.assertFalse(DirectMessage.objects.filter(pk=dm_halfdel2.pk).exists())
        self.assertSetsMessage(resp, 'Message deleted.')

    @override_locale('en')
    def test_IndexView(self):
        self.assertNeedsLogin(reverse('plainui:index'))
        resp = self.client.get(reverse('plainui:index'))
        self.assertEqual(resp.context_data['conf'], self.conf)

    @override_settings(TIME_ZONE='UTC')
    @override_locale('en')
    def test_Fahrplan(self):
        self.conf.start = datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 1, 3, 0, 0, 0, tzinfo=UTC)
        self.conf.save()
        self.user.timezone = 'UTC'
        self.user.save()

        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Some Room', is_public_fahrplan=True)
        room.save()
        track = ConferenceTrack(conference=self.conf, is_public=True, name='Track1')
        track.save()
        event1 = Event(
            conference=self.conf,
            assembly=assembly,
            track=track,
            name='Event1_1',
            is_public=True,
            room=room,
            schedule_start=datetime(2020, 1, 2, 0, 15, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event1.save()
        EventParticipant(is_public=True, participant=self.user, event=event1, role=EventParticipant.Role.SPEAKER).save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly,
            track=track,
            name='Event1_2',
            is_public=True,
            room=room,
            schedule_start=datetime(2020, 1, 2, 1, 15, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
        )
        event2.save()

        self.assertNeedsLogin(reverse('plainui:fahrplan'))
        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'list'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['mode'], 'list')
        self.assertEqual(list(resp.context_data['events']), [event1, event2])
        evt = next(iter(resp.context_data['events']))
        self.assertEqual(evt, event1)
        self.assertEqual(evt.track_name, 'Track1')
        self.assertEqual(evt.speakers[0].speaker_name, self.user.username)

        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'calendar'})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['mode'], 'calendar')
        self.assertEqual(
            resp.context_data['events'],
            {
                'rooms_with_events': [
                    (
                        room,
                        [
                            {'type': 'space', 'minutes': 15.0},
                            {'type': 'event', 'event': event1, 'minutes': 45.0},
                            {'type': 'space', 'minutes': 15.0},
                            {'type': 'event', 'event': event2, 'minutes': 45.0},
                        ],
                    )
                ],
                'calendar_start': datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
                'calendar_end': datetime(2020, 1, 2, 2, 0, 0, tzinfo=UTC),
                'calendar_time_steps': [
                    {'day': 1, 'new_day': True, 'ts': datetime(2020, 1, 2, 0, 0, tzinfo=UTC)},
                    {'day': 1, 'new_day': False, 'ts': datetime(2020, 1, 2, 0, 30, tzinfo=UTC)},
                    {'day': 1, 'new_day': False, 'ts': datetime(2020, 1, 2, 1, 0, tzinfo=UTC)},
                    {'day': 1, 'new_day': False, 'ts': datetime(2020, 1, 2, 1, 30, tzinfo=UTC)},
                ],
                'calendar_step_minutes': 30,
            },
        )

        # some random requests that test building the filter content
        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'calendar', 'show_day_filters': 'y'})
        self.assertEqual(resp.context_data['conf'], self.conf)

        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'calendar', 'show_assembly_filters': 'y'})
        self.assertEqual(resp.context_data['conf'], self.conf)

        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'calendar', 'show_track_filters': 'y'})
        self.assertEqual(resp.context_data['conf'], self.conf)

        # test my fahrplan filter
        # no event on personal schedule
        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'list', 'my': 'y'})
        self.assertEqual(list(resp.context_data['events']), [])

        # event1 on personal schedule
        self.user.favorite_events.add(event1)
        resp = self.client.get(reverse('plainui:fahrplan'), {'mode': 'list', 'my': 'y'})
        self.assertEqual(list(resp.context_data['events']), [event1])

    @override_locale('en')
    def test_PublicFahrplanView(self):
        self.conf.start = datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)
        self.conf.end = datetime(2020, 1, 3, 0, 0, 0, tzinfo=UTC)
        self.conf.save()

        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Some Room', is_public_fahrplan=True)
        room.save()
        room2 = Room(conference=self.conf, assembly=assembly, name='Some Other Room', is_public_fahrplan=False)
        room2.save()
        event = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_1',
            is_public=True,
            room=room,
            schedule_start=datetime(2020, 1, 2, 0, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
            kind=Event.Kind.OFFICIAL,
        )
        event.save()
        event2 = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_2',
            is_public=False,
            room=room,
            schedule_start=datetime(2020, 1, 2, 1, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
            kind=Event.Kind.OFFICIAL,
        )
        event2.save()
        event3 = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event1_3',
            is_public=True,
            room=room2,
            schedule_start=datetime(2020, 1, 2, 2, 0, 0, tzinfo=UTC),
            schedule_duration=timedelta(minutes=45),
            kind=Event.Kind.ASSEMBLY,
        )
        event3.save()

        resp = self.client.get(reverse('plainui:public_fahrplan'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['events']['rooms_with_events'], [(room, [{'type': 'event', 'event': event, 'minutes': 45.0}])])

        self.client.logout()
        resp = self.client.get(reverse('plainui:public_fahrplan'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['events']['rooms_with_events'], [(room, [{'type': 'event', 'event': event, 'minutes': 45.0}])])

    @override_locale('en')
    def test_RoomView(self):
        now = timezone.now()
        self.conf.start = now - timedelta(days=3)
        self.conf.end = now + timedelta(days=3)
        self.conf.save()

        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Room1')
        room.save()
        room_link = RoomLink(room=room, name='stream', link_type=RoomLink.LinkType.VIDEO, link='https://localhost/')
        room_link.save()

        self.assertNeedsLogin(reverse('plainui:room', kwargs={'slug': room.slug}))
        resp = self.client.get(reverse('plainui:room', kwargs={'slug': room.slug}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['room'], room)

    @override_locale('en')
    def test_RoomsView(self):
        now = timezone.now()
        self.conf.start = now - timedelta(days=3)
        self.conf.end = now + timedelta(days=3)
        self.conf.save()

        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Room1')
        room.save()
        room_link = RoomLink(room=room, name='stream', link_type=RoomLink.LinkType.VIDEO)
        room_link.save()

        self.assertNeedsLogin(reverse('plainui:rooms'))
        resp = self.client.get(reverse('plainui:rooms'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['rooms']), [room])

    @override_locale('en')
    def test_UpcomingView(self):
        now = timezone.now()
        self.conf.start = now - timedelta(days=3)
        self.conf.end = now + timedelta(days=3)
        self.conf.save()

        assembly = Assembly(conference=self.conf, slug='assembly1', name='Assembly1', state=Assembly.State.PLACED)
        assembly.save()
        room = Room(conference=self.conf, assembly=assembly, name='Some Room')
        room.save()
        event_pre = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event_pre',
            is_public=True,
            room=room,
            schedule_start=now - timedelta(hours=1),
            schedule_duration=timedelta(minutes=45),
        )
        event_pre.save()
        event_still_running = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event_still_running',
            is_public=True,
            room=room,
            schedule_start=now - timedelta(minutes=30),
            schedule_duration=timedelta(minutes=45),
        )
        event_still_running.save()
        event_soon = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event_soon',
            is_public=True,
            room=room,
            schedule_start=now + timedelta(minutes=15),
            schedule_duration=timedelta(minutes=45),
        )
        event_soon.save()
        event_post = Event(
            conference=self.conf,
            assembly=assembly,
            name='Event_post',
            is_public=True,
            room=room,
            schedule_start=now + timedelta(minutes=35),
            schedule_duration=timedelta(minutes=45),
        )
        event_post.save()

        self.assertNeedsLogin(reverse('plainui:upcoming'))
        resp = self.client.get(reverse('plainui:upcoming'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(list(resp.context_data['events']), [event_still_running, event_soon])

    @override_locale('en')
    def test_LandingView(self):
        self.client.cookies = SimpleCookie()
        resp = self.client.get(reverse('plainui:landing'))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['hide_header'], True)

    @override_settings(DEREFERRER_GLOBAL_ALLOWLIST=[], DEREFERRER_COUNT_ACCESS=[])
    @override_locale('en')
    def _dereferrer_test(self, my_endpoint):
        from django.core.signing import loads

        my_salt = f'{self.user.pk}dereferrer'
        anonymous_salt = '000000dereferrer'

        # local links are always allowed
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': reverse('plainui:landing')}))
        self.assertRedirects(resp, reverse('plainui:landing'))

        # external url shows dereferrer page and the "follow"-link works
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['plain_url'], 'https://http.cat/')
        signed_url = resp.context_data['signed_url']
        self.assertEqual(loads(signed_url, salt=my_salt), 'https://http.cat/')
        resp = self.client.get(reverse('plainui:dereferrer_approved', kwargs={'signed_payload': signed_url}))
        self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)

        # global allowlist works with strings
        with override_settings(DEREFERRER_GLOBAL_ALLOWLIST=['https://http.cat']):
            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
            self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)

            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('http://http.cat/')}))
            self.assertEqual(resp.context_data['plain_url'], 'http://http.cat/')

        # global allowlist works with regexes
        with override_settings(DEREFERRER_GLOBAL_ALLOWLIST=[re.compile(r'https://.*\.cat')]):
            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/301')}))
            self.assertRedirects(resp, 'https://http.cat/301', fetch_redirect_response=False)

            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('http://http.cat/302')}))
            self.assertEqual(resp.context_data['plain_url'], 'http://http.cat/302')

            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('http://http.dog/')}))
            self.assertEqual(resp.context_data['plain_url'], 'http://http.dog/')

        # user can add url to local allow list
        self.assertFalse(UserDereferrerAllowlist.objects.exists())
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
        signed_url = resp.context_data['signed_url']
        resp = self.client.get(reverse('plainui:dereferrer_save', kwargs={'signed_payload': signed_url}))
        self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)
        allowlist_entry = UserDereferrerAllowlist.objects.get()
        self.assertEqual(allowlist_entry.user, self.user)
        self.assertEqual(allowlist_entry.domain, 'https://http.cat')
        self.assertEqual(allowlist_entry.active, True)

        # user allowlist works after adding entry in last request
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
        self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)

        self.assertFalse(DereferrerStats.objects.exists())
        with override_settings(DEREFERRER_COUNT_ACCESS=['https://http.cat']):
            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
            self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)
            stats_object = DereferrerStats.objects.get()
            self.assertEqual(stats_object.domain, 'https://http.cat')
            self.assertEqual(stats_object.hits, 1)

        with override_settings(DEREFERRER_COUNT_ACCESS=[re.compile(r'https://.*\.cat')]):
            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
            self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)
            stats_object.refresh_from_db()
            self.assertEqual(stats_object.hits, 2)

        UserDereferrerAllowlist.objects.all().delete()
        with override_settings(DEREFERRER_COUNT_ACCESS=[re.compile(r'https://.*\.cat')]):
            resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
            signed_url = resp.context_data['signed_url']
            stats_object.refresh_from_db()
            self.assertEqual(stats_object.hits, 2)

            resp = self.client.get(reverse('plainui:dereferrer_approved', kwargs={'signed_payload': signed_url}))
            self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)
            stats_object.refresh_from_db()
            self.assertEqual(stats_object.hits, 3)

            allowlist_entry = UserDereferrerAllowlist(user=self.user, domain='https://http.cat', active=False)
            allowlist_entry.save()
            resp = self.client.get(reverse('plainui:dereferrer_save', kwargs={'signed_payload': signed_url}))
            self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)
            stats_object.refresh_from_db()
            self.assertEqual(stats_object.hits, 4)
            allowlist_entry.refresh_from_db()
            self.assertTrue(allowlist_entry.active)

        # prevents loops
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote(reverse(my_endpoint, kwargs={'dst': '/foo'}))}))
        self.assertRedirects(resp, reverse('plainui:index'))

        # works with anonymous user
        self.client.cookies = SimpleCookie()
        # local links are always allowed
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': reverse('plainui:landing')}))
        self.assertRedirects(resp, reverse('plainui:landing'))

        # external url shows dereferrer page and the "follow"-link works
        resp = self.client.get(reverse(my_endpoint, kwargs={'dst': quote('https://http.cat/')}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['plain_url'], 'https://http.cat/')
        signed_url = resp.context_data['signed_url']
        self.assertEqual(loads(signed_url, salt=anonymous_salt), 'https://http.cat/')
        resp = self.client.get(reverse('plainui:dereferrer_approved', kwargs={'signed_payload': signed_url}))
        self.assertRedirects(resp, 'https://http.cat/', fetch_redirect_response=False)

    @override_locale('en')
    def test_DereferrerView(self):
        self._dereferrer_test('plainui:dereferrer')

    @override_settings(RATELIMIT_ENABLE=False)
    @override_locale('en')
    def test_ReportContentView(self):
        from plainui.forms import REPORT_CATEGORIES

        category_id = iter(REPORT_CATEGORIES.keys()).__next__()

        user2 = PlatformUser(username='testuser2')
        user2.save()
        dm_recv = DirectMessage(conference=self.conf, sender=user2, recipient=self.user, subject='sj1', body='bd1')
        dm_recv.save()
        bbe = BulletinBoardEntry(conference=self.conf, owner=self.user, title='board entry title', text='board entry text')
        bbe.save()

        self.assertNeedsLogin(reverse('plainui:report_content'), check_user=True)
        resp = self.client.get(reverse('plainui:report_content'), {'kind': 'url', 'kind_data': reverse('plainui:index')})
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['form']['kind'].value(), 'url')
        self.assertEqual(resp.context_data['form']['kind_data'].value(), reverse('plainui:index'))

        resp = self.client.post(
            reverse('plainui:report_content'),
            {
                'kind': 'url',
                'kind_data': reverse('plainui:index'),
                'next': '',
                'category': category_id,
                'message': 'Some Message',
                'message2': 'Some Resolution',
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertEqual(len(mail.outbox), 1)
        sent_mail = mail.outbox[0]
        self.assertEqual(sent_mail.to, [REPORT_CATEGORIES[category_id][2]])
        self.assertEqual(sent_mail.subject, f'New {REPORT_CATEGORIES[category_id][1]} Report')
        self.assertMultiLineEqual(
            (
                f"testuser ({self.user.id!s}) sent a new {REPORT_CATEGORIES[category_id][1]} report.\n"
                f"Profile of the Report Author: {hub_absolute('plainui:user_by_uuid', uuid=str(self.user.uuid))}\n"
                f"\n"
                f"\n"
                f"--- Problem description ------------------------------------------------------\n"
                f"Some Message\n"
                f"\n"
                f"--- Solution description -----------------------------------------------------\n"
                f"Some Resolution\n"
                f"\n"
                f"--- Problematic content ------------------------------------------------------\n"
                f"{reverse('plainui:index')}\n"
            ),
            sent_mail.body,
        )
        mail.outbox.clear()

        resp = self.client.post(
            reverse('plainui:report_content'),
            {
                'kind': 'pn',
                'kind_data': str(dm_recv.pk),
                'next': reverse('plainui:personal_message_show', kwargs={'msg_id': str(dm_recv.pk)}),
                'category': category_id,
                'message': 'Some Message',
                'message2': 'Some Resolution',
            },
        )
        self.assertRedirects(resp, reverse('plainui:personal_message_show', kwargs={'msg_id': str(dm_recv.pk)}))
        self.assertEqual(len(mail.outbox), 1)
        sent_mail = mail.outbox[0]
        self.assertEqual(sent_mail.to, [REPORT_CATEGORIES[category_id][2]])
        self.assertEqual(sent_mail.subject, f'New {REPORT_CATEGORIES[category_id][1]} Report')
        self.maxDiff = None
        self.assertMultiLineEqual(
            sent_mail.body,
            (
                f'testuser ({self.user.id!s}) sent a new {REPORT_CATEGORIES[category_id][1]} report.\n'
                f"Profile of the Report Author: {reverse('plainui:user_by_uuid', kwargs={'uuid': str(self.user.uuid)})}\n"
                f'\n\n'
                f"--- Problem description ------------------------------------------------------\n"
                f'Some Message\n'
                f'\n'
                f"--- Solution description -----------------------------------------------------\n"
                f"Some Resolution\n"
                f"\n"
                f"--- Problematic content ------------------------------------------------------\n"
                f'DM by testuser2 ({user2.id!s}) to testuser ({self.user.id!s}) at {localize(localtime(dm_recv.timestamp))} ({dm_recv.pk!s}).\n'
                f'Subject: "sj1"\n'
                f'Message: "bd1"\n'
            ),
        )
        mail.outbox.clear()

        resp = self.client.post(
            reverse('plainui:report_content'),
            {
                'kind': 'board',
                'kind_data': str(bbe.pk),
                'next': '',
                'category': category_id,
                'message': 'Other Message',
                'message2': 'Other Resolution',
            },
        )
        self.assertRedirects(resp, reverse('plainui:index'))
        self.assertEqual(len(mail.outbox), 1)
        sent_mail = mail.outbox[0]
        self.assertEqual(sent_mail.to, [REPORT_CATEGORIES[category_id][2]])
        self.assertEqual(sent_mail.subject, f'New {REPORT_CATEGORIES[category_id][1]} Report')
        self.assertMultiLineEqual(
            sent_mail.body,
            (
                f'testuser ({self.user.id!s}) sent a new {REPORT_CATEGORIES[category_id][1]} report.\n'
                f"Profile of the Report Author: {reverse('plainui:user_by_uuid', kwargs={'uuid': str(self.user.uuid)})}\n"
                f"MODERATION: {reverse('backoffice:moderation-board-detail', kwargs={'pk': bbe.pk})}\n"
                f'\n'
                f'--- Problem description ------------------------------------------------------\n'
                f'Other Message\n'
                f'\n'
                f'--- Solution description -----------------------------------------------------\n'
                f'Other Resolution\n'
                f'\n'
                f'--- Problematic content ------------------------------------------------------\n'
                f'Board Entry by testuser ({self.user.id!s}) at {localize(localtime(dm_recv.timestamp))} ({bbe.pk!s})\n'
                f'Title: "board entry title"\n'
                f'Text: "board entry text"\n'
            ),
        )
        mail.outbox.clear()

        # testing the rate limiting: Will block after 1 allowed requests
        with override_settings(RATELIMIT_ENABLE=True):
            resp = self.client.post(
                reverse('plainui:report_content'),
                {
                    'kind': 'board',
                    'kind_data': str(bbe.pk),
                    'next': '',
                    'category': category_id,
                    'message': 'Other Message',
                    'message2': 'Other Resolution',
                },
            )
            self.assertRedirects(resp, reverse('plainui:index'))
            self.assertEqual(len(mail.outbox), 1)
            mail.outbox.clear()

            resp = self.client.post(
                reverse('plainui:report_content'),
                {
                    'kind': 'board',
                    'kind_data': str(bbe.pk),
                    'next': '',
                    'category': category_id,
                    'message': 'Other Message',
                    'message2': 'Other Resolution',
                },
            )
            self.assertIsInstance(resp, HttpResponseRateLimited)
            self.assertEqual(len(mail.outbox), 0)

    @override_locale('en')
    def test_UserView(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()
        cm = ConferenceMember(conference=self.conf, user=user2, description_en='descr')
        cm.save()
        assembly = Assembly(conference=self.conf, slug='s', name='a')
        assembly.save()
        badge1 = Badge(conference=self.conf, name='badge-name', issuing_assembly=assembly)
        badge1.save()
        badge2 = Badge(conference=self.conf, name='hidden-badge', issuing_assembly=assembly)
        badge2.save()
        user_badge1 = UserBadge(user=user2, badge=badge1, accepted_by_user=True, visibility=UserBadgeVisibility.PUBLIC)
        user_badge1.save()
        user_badge1.acceptByUser()
        user_badge2 = UserBadge(user=user2, badge=badge2)
        user_badge2.save()

        self.assertNeedsLogin(reverse('plainui:user', kwargs={'user_slug': str(user2.slug)}))
        resp = self.client.get(reverse('plainui:user', kwargs={'user_slug': str(user2.slug)}))
        self.assertEqual(resp.context_data['conf'], self.conf)
        self.assertEqual(resp.context_data['display_user'], user2)
        self.assertEqual(resp.context_data['description_html'], render_markdown(self.conf, 'descr'))
        self.assertEqual(list(resp.context_data['badges']), [user_badge1])

    @override_locale('en')
    def test_UserByUuidView(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()

        self.assertNeedsLogin(reverse('plainui:user_by_uuid', kwargs={'uuid': str(user2.uuid)}))
        resp = self.client.get(reverse('plainui:user_by_uuid', kwargs={'uuid': str(user2.uuid)}))
        self.assertRedirects(resp, reverse('plainui:user', kwargs={'user_slug': str(user2.slug)}))

    @override_settings(RATELIMIT_ENABLE=False)
    def test_badge_redeem_view_get(self):
        user2 = PlatformUser(username='testuser2')
        user2.save()
        cm = ConferenceMember(conference=self.conf, user=user2, has_ticket=True, description_en='descr')
        cm.save()
        assembly = Assembly(conference=self.conf, slug='s', name='a')
        assembly.save()
        badge = Badge(conference=self.conf, name='badge-name', issuing_assembly=assembly, state=Badge.State.PUBLIC)
        badge.save()
        badge_token = BadgeToken.objects.create(badge=badge)
        resp = self.client.get(reverse('plainui:manage_badges'), {'redeem_token': badge_token.token})
        self.assertRedirects(resp, reverse('plainui:badge_award', kwargs={'pk': badge.pk}))
        self.assertTrue(UserBadge.objects.filter(user=self.user, badge=badge).exists())
        self.client.force_login(user2)
        resp = self.client.get(reverse('plainui:badge_award', kwargs={'pk': badge.pk}))
        self.assertRedirects(resp, reverse('plainui:badge', kwargs={'pk': badge.pk}))

    @override_settings(RATELIMIT_ENABLE=False)
    def test_badge_redeem_view_post(self):
        assembly = Assembly(conference=self.conf, slug='s', name='a')
        assembly.save()
        badge = Badge(conference=self.conf, name='badge-name', issuing_assembly=assembly)
        badge.save()
        badge_token = BadgeToken.objects.create(badge=badge)
        resp = self.client.post(reverse('plainui:manage_badges'), {'token': badge_token.token, 'purpose': 'redeem_token'})
        self.assertRedirects(resp, reverse('plainui:badge_award', kwargs={'pk': badge.pk}))
        self.assertTrue(UserBadge.objects.filter(user=self.user, badge=badge).exists())

    def test_badge_redeem_view_rate_limit_post(self):
        assembly = Assembly(conference=self.conf, slug='s', name='a')
        assembly.save()

        resp = self.client.post(reverse('plainui:manage_badges'), {'token': '12345678901234567890', 'purpose': 'redeem_token'})
        self.assertEqual(resp.status_code, 200)
        resp = self.client.post(reverse('plainui:manage_badges'), {'token': '12345678901234567892', 'purpose': 'redeem_token'})
        self.assertEqual(resp.status_code, 429)
        self.assertIsInstance(resp, HttpResponseRateLimited)

    def test_badge_redeem_view_rate_limit_get(self):
        assembly = Assembly(conference=self.conf, slug='s', name='a')
        assembly.save()

        resp = self.client.get(reverse('plainui:manage_badges'), {'redeem_token': '12345678901234567890'})
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get(reverse('plainui:manage_badges'), {'redeem_token': '12345678901234567892'})
        self.assertEqual(resp.status_code, 429)
        self.assertIsInstance(resp, HttpResponseRateLimited)


class TestConferenceDetection(TestCase):
    def setUp(self):
        self.request = RequestFactory().get('/')
        self.view = LandingView()

    def test_no_conference_config(self):
        with self.assertRaisesMessage(ImproperlyConfigured, expected_message='No conference found in database!'):
            self.view.dispatch(self.request)

    @override_settings(SELECTED_CONFERENCE_ID=None)
    def test_valid_conference_config(self):
        self.conf = Conference(id=TEST_CONF_ID, name='conf_asdf', slug='slug1')
        self.conf.save()
        self.view.dispatch(self.request)

    def setUpTwoDatabases(self):
        self.conf = Conference(id=TEST_CONF_ID, name='slug1', slug='slug1')
        self.conf.save()
        self.conf2 = Conference(id=TEST_CONF_ID_2, name='slug2', slug='slug2')
        self.conf2.save()

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
    def test_valid_conference(self):
        self.setUpTwoDatabases()

        self.view.dispatch(self.request)

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID_3)
    def test_invalid_conference_config_wrong_setting(self):
        self.setUpTwoDatabases()

        with self.assertRaisesMessage(ImproperlyConfigured, expected_message='SELECTED_CONFERENCE_ID from settings not found in database!'):
            self.view.dispatch(self.request)

    @override_settings(SELECTED_CONFERENCE_ID=None)
    def test_invalid_conference_config_no_setting(self):
        self.setUpTwoDatabases()

        with self.assertRaisesMessage(ImproperlyConfigured, expected_message='Multiple Conferences found, set SELECTED_CONFERENCE_ID in settings!'):
            self.view.dispatch(self.request)
