__all__ = (
    'DereferrerAddToAllowlistView',
    'DereferrerRedirectView',
    'DereferrerRemoveFromAllowlistView',
    'DereferrerView',
)

from urllib.parse import ParseResult, unquote, urlparse

from django.conf import settings
from django.contrib import messages
from django.core import signing
from django.core.exceptions import PermissionDenied
from django.db import connection
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext
from django.views.generic.base import View

from core.models import (
    DereferrerStats,
    UserDereferrerAllowlist,
)
from core.utils import url_in_allowlist

from plainui.views.utils import ConferenceRequiredMixin


class BaseDereferrerView(ConferenceRequiredMixin, View):
    only_published = False

    def gen_salt(self):
        if self.request.user.is_authenticated:
            return f'{self.request.user.pk}dereferrer'
        else:
            return '000000dereferrer'

    def unsign_payload(self, signed_payload: str) -> str:
        try:
            return signing.loads(signed_payload, salt=self.gen_salt())
        except signing.BadSignature:
            raise PermissionDenied

    def sign_payload(self, payload: str) -> str:
        return signing.dumps(payload, salt=self.gen_salt())

    def increment_counter_and_redirect(self, url: ParseResult):
        scheme = url.scheme if url.scheme else self.request.scheme
        scheme_and_netloc = scheme + '://' + url.netloc

        if url_in_allowlist(scheme_and_netloc, settings.DEREFERRER_COUNT_ACCESS):
            with connection.cursor() as cursor:
                cursor.execute(
                    'INSERT INTO '
                    + DereferrerStats._meta.db_table
                    + ' AS ds (domain, hits) VALUES (%s, 1) \
                        ON CONFLICT (domain) DO UPDATE SET hits = ds.hits + 1',
                    [scheme_and_netloc],
                )

        return HttpResponseRedirect(url.geturl())


class DereferrerRedirectView(BaseDereferrerView):
    require_login = False
    require_conference_member = False

    def get(self, request, signed_payload, *args, **kwargs):
        raw_url = self.unsign_payload(signed_payload)
        url = urlparse(raw_url)

        return self.increment_counter_and_redirect(url)


class DereferrerAddToAllowlistView(BaseDereferrerView):
    require_user = True

    def get(self, request, signed_payload, *args, **kwargs):
        raw_url = self.unsign_payload(signed_payload)
        url = urlparse(raw_url)

        scheme = url.scheme if url.scheme else self.request.scheme
        scheme_and_netloc = scheme + '://' + url.netloc

        # add to the user's allowlist
        UserDereferrerAllowlist.objects.update_or_create(domain=scheme_and_netloc, user=request.user, defaults={'active': True})

        return self.increment_counter_and_redirect(url)


class DereferrerView(BaseDereferrerView):
    require_login = False
    require_conference_member = False
    template_name = 'plainui/dereferrer.html.j2'

    def get(self, request, dst, *args, **kwargs):
        destination = unquote(dst)
        url = urlparse(destination)

        # local urls are always allowed
        allowlist_match = not url.scheme and not url.netloc
        scheme = url.scheme if url.scheme else request.scheme
        scheme_and_netloc = scheme + '://' + url.netloc

        # prevent loops
        my_endpoint = reverse('plainui:dereferrer', kwargs={'dst': '/'})[:-1]
        if allowlist_match and (destination.startswith(my_endpoint) or not destination.startswith('/')):
            return redirect(reverse('plainui:index'))

        # is the url allowed via global allow list?
        if not allowlist_match:
            allowlist_match = url_in_allowlist(scheme_and_netloc, settings.DEREFERRER_GLOBAL_ALLOWLIST)

        # is the url allowed via the user's allowlist?
        if not allowlist_match and request.user.is_authenticated:
            allowlist_match = UserDereferrerAllowlist.objects.filter(domain=scheme_and_netloc, user=request.user, active=True).exists()

        # if the url is allowed, redirect directly
        if allowlist_match:
            return self.increment_counter_and_redirect(url)
        else:
            # if the url is not allowed, show the dereferrer page
            signed_url = self.sign_payload(url.geturl())
            return TemplateResponse(
                request,
                self.template_name,
                context={
                    'conf': self.conf,
                    'plain_url': url.geturl(),
                    'domain': scheme_and_netloc,
                    'signed_url': signed_url,
                    'can_allow': request.user.is_authenticated,
                    'disable_share': True,
                },
            )


class DereferrerRemoveFromAllowlistView(View):
    def get(self, request, *args, **kwargs):
        return redirect(reverse('plainui:userprofile'))

    def post(self, request, *args, **kwargs):
        try:
            entry = UserDereferrerAllowlist.objects.get(pk=request.POST['pk'])
        except UserDereferrerAllowlist.DoesNotExist:
            return redirect(reverse('plainui:userprofile'))

        entry.delete()
        messages.success(request, gettext('UserDereferrerAllowlist--deleted'))
        return redirect(reverse('plainui:userprofile'))
