__all__ = (
    'FahrplanView',
    'PublicFahrplanView',
)

from datetime import datetime

from django.http import Http404
from django.utils import timezone
from django.views.generic.base import TemplateView

from core.models import (
    Assembly,
    ConferenceTrack,
    Event,
)
from core.utils import str2bool

from plainui.views.utils import ConferenceRequiredMixin, event_filter, organize_events_for_calendar, session_get_favorite_events


class FahrplanView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/fahrplan.html.j2'
    filter_opts = {}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        timezone_warning = timezone.get_current_timezone() != self.conf.timezone
        context['timezone_warning'] = timezone_warning
        if timezone_warning:
            current_timezone = timezone.get_current_timezone()
            if current_timezone is not None:
                context['current_timezone'] = current_timezone.key if current_timezone.key else datetime.now(current_timezone).tzname()
                context['conference_timezone'] = self.conf.timezone

        context['my_favorite_events'] = session_get_favorite_events(self.request.session, self.request.user)

        if self.conf.days_count == 0:
            raise Http404

        context['days'] = self.conf.days
        mode = self.request.GET.get('mode', 'calendar')
        day = self.request.GET.get('day', None)
        kind = self.request.GET.get('kind', 'official')
        track = self.request.GET.get('track', None)
        assembly = self.request.GET.get('assembly', None)
        show_day_filters = self.request.GET.get('show_day_filters') == 'y'
        show_assembly_filters = self.request.GET.get('show_assembly_filters') == 'y'
        show_track_filters = self.request.GET.get('show_track_filters') == 'y'
        my_fahrplan = self.request.GET.get('my') == 'y'
        is_recorded = str2bool(self.request.GET.get('rec', ''), allow_null=True)

        to_set = self.request.GET.get('set', None)
        if to_set:
            if to_set[0] == 'k':
                kind = to_set[1:]
                if kind not in {'all', 'official', 'assembly', 'sos'}:
                    kind = 'official'
            elif to_set[0] == 'a':
                assembly = to_set[1:]
            elif to_set[0] == 'd':
                day = to_set[1:]
            elif to_set[0] == 'f':
                f = to_set[1:]
                if f == 'day':
                    show_day_filters = not show_day_filters
                elif f == 'assembly':
                    show_assembly_filters = not show_assembly_filters
                elif f == 'track':
                    show_track_filters = not show_track_filters
                elif f == 'my':
                    my_fahrplan = not my_fahrplan
            elif to_set[0] == 'm':
                mode = to_set[1:]
            elif to_set[0] == 't':
                track = to_set[1:]
            elif to_set[0] == 'r':
                match to_set[1:]:
                    case '1' | 'y':
                        is_recorded = True
                    case '0' | 'n':
                        is_recorded = False
                    case _:
                        is_recorded = None

        context['show_day_filters'] = show_day_filters
        context['show_assembly_filters'] = show_assembly_filters
        context['show_track_filters'] = show_track_filters

        context['my_fahrplan'] = my_fahrplan
        self.filter_opts['user_schedule_only'] = my_fahrplan

        if day:
            try:
                day = int(day)
            except ValueError:
                day = None
        else:
            day = None
        context['day'] = self.conf.day(day) if day else None

        if kind == 'all':
            public_fahrplan = None
            kinds = None
        elif kind == 'sos':
            public_fahrplan = None
            kinds = [Event.Kind.SELF_ORGANIZED]
        elif kind == 'assembly':
            public_fahrplan = False
            kinds = [Event.Kind.ASSEMBLY]
        else:
            public_fahrplan = True
            kinds = None
        context['kind'] = kind

        if kind == 'sos':
            mode = 'list'
        context['calendar_mode_disabled'] = kind == 'sos'

        if show_assembly_filters:
            # TODO show only assemblies that have official / community events
            assemblies = Assembly.objects.conference_accessible(self.conf).filter(is_official=False, events__isnull=False).distinct()
            context['assemblies'] = assemblies
            assembly = assemblies.get(slug=assembly) if assembly else None
            context['assembly'] = assembly
        else:
            context['assembly'] = assembly = None

        tracks = ConferenceTrack.objects.filter(conference=self.conf, is_public=True)
        context['tracks'] = tracks
        track = tracks.get(slug=track) if track else None
        context['track'] = track

        context['is_recorded'] = is_recorded

        events = event_filter(
            self.request.user,
            self.conf,
            day=day,
            kinds=kinds,
            assembly=assembly,
            track=track,
            calendar_mode=mode == 'calendar',
            public_fahrplan=public_fahrplan,
            is_recorded=is_recorded,
            **self.filter_opts,
        )

        if mode == 'calendar':
            context['mode'] = 'calendar'
            events = organize_events_for_calendar(self.conf, events)
            context['events'] = events
        elif mode == 'list':
            context['mode'] = 'list'
            context['events'] = events
        return context


class PublicFahrplanView(ConferenceRequiredMixin, TemplateView):
    require_conference_member = False
    require_login = False
    template_name = 'plainui/public_fahrplan.html.j2'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(conf=self.conf, **kwargs)
        context['mode'] = 'calendar'

        # TODO: Timezone warning macht im public fahrplan keinen Sinn, wir stirppen momentan die Sessions im Loadbalancer.
        #       Da sollten wir immer in der offiziellen Zeit anzeigen. Eventuell mit dauerhaftem Timezone-Hinweis?
        timezone_warning = self.conf.timezone and timezone.get_current_timezone() != self.conf.timezone
        context['timezone_warning'] = timezone_warning
        if timezone_warning:
            current_timezone = timezone.get_current_timezone()
            if current_timezone is not None:
                context['current_timezone'] = current_timezone.key if current_timezone.key else datetime.now(current_timezone).tzname()
                context['conference_timezone'] = self.conf.timezone

        events = event_filter(self.request.user, self.conf, public_fahrplan=True)
        context['events'] = organize_events_for_calendar(self.conf, events)

        tracks = ConferenceTrack.objects.filter(conference=self.conf, is_public=True)
        context['tracks'] = tracks

        return context
