__all__ = (
    'ProjectView',
    'ProjectsListView',
)


from typing import Any

from django.db.models import Prefetch, QuerySet
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.views.generic import DetailView

from core.models import (
    Project,
    TagItem,
)
from core.views.list_views import AlphabeticalPaginatorView, PrefetchTagsMixin

from plainui.views.utils import (
    ConferenceRequiredMixin,
    session_get_favorite_projects,
)


class ProjectView(ConferenceRequiredMixin, DetailView):
    model = Project
    template_name = 'plainui/projects/detail.html.j2'
    context_object_name = 'project'
    self_organized = False

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.project = self.object = self.get_object()
        if self.self_organized and not self.project.is_self_organized:
            return HttpResponseRedirect(reverse('plainui:project', kwargs={'slug': self.project.slug}))
        if not self.self_organized and self.project.is_self_organized:
            return HttpResponseRedirect(reverse('plainui:so_project', kwargs={'slug': self.project.slug}))
        return super().get(request, *args, **kwargs)

    def get_queryset(self, queryset: QuerySet | None = None) -> QuerySet[Project]:
        return (
            Project.objects.conference_accessible(self.conf)
            .select_related('owner', 'assembly')
            .prefetch_related('links')
            .prefetch_related(
                Prefetch('tags', to_attr='prefetched_tags', queryset=TagItem.objects.select_related('tag').filter(tag__is_public=True)),
            )
        )

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'self_organized': self.self_organized,
            'report_info': {'lookup_key': self.project.slug},
            'is_favorite': str(self.object.id) in session_get_favorite_projects(self.request.session, self.request.user),
        }


class ProjectsListView(ConferenceRequiredMixin, PrefetchTagsMixin, AlphabeticalPaginatorView):
    model = Project
    template_name = 'plainui/projects/list.html.j2'
    context_object_name = 'projects'
    filters = {
        'name': ('name__icontains', 'str'),
    }
    ordering = ['name']

    def get_queryset(self, queryset: QuerySet | None = None) -> QuerySet[Project]:
        qs = Project.objects.conference_accessible(self.conf)
        get_parameters = getattr(self.request, 'GET', {})
        if project_type := get_parameters.get('type'):
            self.params['type'] = project_type
            match project_type.lower():
                case 'sop':
                    qs = qs.filter(owner__isnull=False)
                case 'assembly':
                    qs = qs.filter(assembly__isnull=False)
                case _:
                    qs = qs.none()
        return self.add_tag_annotations(super().get_queryset(qs).select_related('owner', 'assembly').prefetch_related('tags'))

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'report_info': {'enabled': True},
            'is_favorite_projects': session_get_favorite_projects(self.request.session, self.request.user),
        }
